/*
 * This file is part of OpenTTD.
 * OpenTTD is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, version 2.
 * OpenTTD is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details. You should have received a copy of the GNU General Public License along with OpenTTD. If not, see <http://www.gnu.org/licenses/>.
 */

/** @file command_func.h Functions related to commands. */



#ifndef COMMAND_FUNC_H
#define COMMAND_FUNC_H



#include "command_type.h"
#include "company_type.h"



/**
 * Define a default return value for a failed command.
 *
 * This variable contains a CommandCost object with is declared as "failed".
 * Other functions just need to return this error if there is an error,
 * which doesn't need to specific by a StringID.
 */
static const CommandCost CMD_ERROR = CommandCost(INVALID_STRING_ID);



/**
 * Returns from a function with a specific StringID as error.
 *
 * This macro is used to return from a function. The parameter contains the
 * StringID which will be returned.
 *
 * @param errcode The StringID to return
 */
#define return_cmd_error(errcode) return CommandCost(errcode);



extern Money _additional_cash_required;



CommandCost  ProcessCmd                  (CommandContainer container);
bool         ProcessCmdP                 (CommandContainer container);
CommandCost  ProcessCmdPInternal         (CommandContainer container);
void         NetworkSendCommand          (CommandContainer container);

bool         IsValidCommand              (uint32 cmd);
CommandFlags GetCommandFlags             (uint32 cmd);
const char   *GetCommandName             (uint32 cmd);
Money        GetAvailableMoneyForCommand ();
bool         IsCommandAllowedWhilePaused (uint32 cmd);



/*** All command callbacks that exist ***/

/* ai/ai_instance.cpp */
CommandCallback CcAI;


/* airport_gui.cpp */
CommandCallback CcBuildAirport;


/* bridge_gui.cpp */
CommandCallback CcBuildBridge;


/* dock_gui.cpp */
CommandCallback CcBuildDocks;
CommandCallback CcPlaySound_CONSTRUCTION_WATER;


/* depot_gui.cpp */
CommandCallback CcCloneVehicle;


/* game/game_instance.cpp */
CommandCallback CcGame;


/* group_gui.cpp */
CommandCallback CcCreateGroup;
CommandCallback CcAddVehicleNewGroup;


/* industry_gui.cpp */
CommandCallback CcBuildIndustry;


/* main_gui.cpp */
CommandCallback CcPlaySound_EXPLOSION;
CommandCallback CcPlaceSign;
CommandCallback CcTerraform;


/* rail_gui.cpp */
CommandCallback CcPlaySound_CONSTRUCTION_RAIL;
CommandCallback CcRailDepot;
CommandCallback CcStation;
CommandCallback CcBuildRailTunnel;


/* road_gui.cpp */
CommandCallback CcPlaySound_CONSTRUCTION_OTHER;
CommandCallback CcBuildRoadTunnel;
CommandCallback CcRoadDepot;
CommandCallback CcRoadStop;


/* train_gui.cpp */
CommandCallback CcBuildWagon;


/* town_gui.cpp */
CommandCallback CcFoundTown;
CommandCallback CcFoundRandomTown;


/* vehicle_gui.cpp */
CommandCallback CcBuildPrimaryVehicle;
CommandCallback CcStartStopVehicle;



ProcessCmdFlag CommandFlagsToDCFlags(CommandFlags cmd_flags);



#endif /* COMMAND_FUNC_H */
