/*
 * This file is part of OpenTTD.
 * OpenTTD is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, version 2.
 * OpenTTD is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details. You should have received a copy of the GNU General Public License along with OpenTTD. If not, see <http://www.gnu.org/licenses/>.
 */

/** @file vector.c Vector Functions. */

#include <stdio.h>
#include <stdlib.h>

#include "vector_func.h"

/* begin static */

static void VectorResize(Vector *v, int amount);

/* end static */

void VectorInit(Vector *v)
{
    v->size = 2;
    v->in_use = 0;
    v->items = malloc(sizeof(void *) * v->size);
}

int VectorInUse(Vector *v)
{
    return v->in_use;
}

void VectorAdd(Vector *v, void *item)
{
    if (v->size == v->in_use) {
        VectorResize(v, 1);
	}
    v->items[v->in_use++] = item;
}

void VectorInsert(Vector *v, int index, void *item)
{
    if (v->size == v->in_use) {
        VectorResize(v, 1);
	}

    for (int i = v->size-1; i > index; --i) {
        v->items[i] = v->items[i - 1];
	}

    v->items[index] = item;
}

void VectorSet(Vector *v, int index, void *item)
{
    if (index >= 0 && index < v->in_use) {
        v->items[index] = item;
	}
}

void *VectorFetch(Vector *v, int index)
{
    if (index >= 0 && index < v->in_use) {
        return v->items[index];
	}
    return NULL;
}

void VectorDelete(Vector *v, int index)
{
    if (index < 0 || index >= v->in_use) {
        return;
	}

    v->items[index] = NULL;

    for (int i = index; i < v->in_use - 1; ++i) {
        v->items[i] = v->items[i + 1];
        v->items[i + 1] = NULL;
    }

    --v->in_use;

    if (v->in_use > 0 && v->in_use == v->size - 1) {
        VectorResize(v, v->size - 1);
	}
}

static void VectorResize(Vector *v, int amount)
{
    void *items = realloc(v->items, sizeof(void *) * (v->size + amount));
    if (items) {
        v->items = items;
        v->size += amount;
    }
}

void VectorFree(Vector *v)
{
    free(v->items);
}
