/*
 * This file is part of OpenTTD.
 * OpenTTD is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, version 2.
 * OpenTTD is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details. You should have received a copy of the GNU General Public License along with OpenTTD. If not, see <http://www.gnu.org/licenses/>.
 */

/** @file date_type.h Types related to the dates in OpenTTD. */

#ifndef DATE_TYPE_H
#define DATE_TYPE_H

#include "stdafx.h"

typedef int32 Date;      ///< The type to store our dates in
typedef int32 DateFract; ///< The fraction of a date we're in, i.e. the number of ticks since the last date changeover
typedef int64 Ticks;     ///< The type to store ticks in
typedef int64 TimeTick;  ///< The type to use for time related ticks.
typedef int32 Year;      ///< Type for the year, note: 0 based, i.e. starts at the year 0.
typedef uint8 Month;     ///< Type for the month, note: 0 based, i.e. 0 = January, 11 = December.
typedef uint8 Day;       ///< Type for the day of the month, note: 1 based, first day of a month is 1.
typedef uint8 Hour;      ///< Type for the hour of the day, note: 24 hour based.
typedef uint8 Minute;    ///< Type for the minute of the hour.

extern Year      _cur_year       ;    ///< Current year, starting at 0.
extern Month     _cur_month      ;    ///< Current month (0..11).
extern Day       _cur_day        ;    ///< Current day of the month (1..31).
extern Hour      _cur_hour       ;    ///< Current hour of the day.
extern Minute    _cur_minute     ;    ///< Cuurent minute of the day.
extern Date      _date           ;    ///< Current date in days (day counter).
extern DateFract _date_fract     ;    ///< Fractional part of the day.
extern uint16    _tick_counter   ;    ///< Ever incrementing (and sometimes wrapping) tick counter for setting off various events.
extern uint8     _scale          ;   ///< The scale to modify objects with.
extern uint8     _reverse_scale  ;    ///< The reversed scale to modify objects with.
extern int32     _time_factor    ; ///< Factor multiplying TIME_UNIT with.
extern int32     _tick_value     ;   ///< The value of each _date_fract tick for independent time storage.
extern uint16    _minute_counter ;    ///< Counter for updating time variables.
extern TimeTick  _day_tick       ;    ///< The absolute tick of the day
extern TimeTick  _time_tick      ;    ///< The absolute tick of time.
extern TimeTick  _tick_loop      ;    ///< Independent tick loop for non-fixed time periodic triggers.
extern bool      _enable_clock   ;    ///< Whether or not to enable display of a clock, depending on the time factor.

/** Available settings for _time_factor and associated tick value for time storage.
 *  Tick value is MAX_TIME_FACTOR / _time_factor with the value for largest factor set to 1.
 *  Set MAX_TIME_FACTOR in date_type.cpp to the highest value in the array when changing this.
 */
static const int32 _time_factor_options[] = {
	    1, ///< multiplier 1, no change
	    2, ///< multiplier
	    4, ///< multiplier
	    6, ///< multiplier
	   10, ///< multiplier
	   30, ///< multiplier
	   60, ///< multiplier
	  120, ///< multiplier
	  360, ///< multiplier
	  720, ///< multiplier
	 1440, ///< multiplier, default
	 2880, ///< multiplier
	 7200, ///< multiplier
	14400, ///< multiplier
	43200  ///< multiplier
};

static const int32 _tick_value_options[] = {
	43200, ///< tick value
	21600, ///< tick value
	10800, ///< tick value
	 7200, ///< tick value
	 4320, ///< tick value
	 1440, ///< tick value
	  720, ///< tick value
	  360, ///< tick value
	  120, ///< tick value
	   60, ///< tick value
	   30, ///< tick value, default
	   15, ///< tick value
	    6, ///< tick value
	    3, ///< tick value
	    1  ///< tick value
};

/** Available settings for autosave intervals. */
static const Month _autosave_months[] = {
	 0, ///< never
	 1, ///< every month
	 3, ///< every 3 months
	 6, ///< every 6 months
	12, ///< every 12 months
};

static const Day _autosave_days[] = {
	 0, ///< never
	 1, ///< every day
	 7, ///< every 7 days
	14, ///< every 14 days
	28, ///< every 28 days
};

/**
 * Constant value definitions.
 * DAY_TICKS used to be 74 (int) with one game day lasting 2.22 seconds (1 tick ~ 30 ms).
 * Now it's TIME_UNIT and 60 ticks to align better with real time for real clock use.
 */
static const int64 TIME_UNIT               = 60;                        ///< Minimal time unit in use.
static const int64 MAX_TIME_FACTOR         = 43200;                     ///< Maximum allowed day length factor.
static const int64 DAY_LENGTH              = 2592000;                   ///< Day length in absolute ticks.
static const int64 HOUR                    = 108000;                    ///< One hour in absolute ticks.
static const int64 MINUTE                  = 1800;                      ///< One minute in absolute ticks.
static const int   DAYS_IN_YEAR            = 365;                       ///< days per year
static const int   DAYS_IN_LEAP_YEAR       = 366;                       ///< sometimes, you need one day more...
static const int   MONTHS_IN_YEAR          = 12;                        ///< months per year

/* Interval definitions. */
static const int64 CARGO_AGING_TICKS       = DAY_LENGTH +  9;           ///< cycle duration for aging cargo
static const int64 TOWN_GROWTH_TICKS       = DAY_LENGTH + 51;           ///< cycle duration for towns trying to grow.
#define            INDUSTRY_PRODUCE_TICKS  (DAY_LENGTH / _scale)        ///< cycle duration for industry production
#define            INDUSTRY_CUT_TREE_TICKS (INDUSTRY_PRODUCE_TICKS * 2) ///< cycle duration for lumber mill's extra action

/* Other definitions. */
static const int8  INVALID                 = -1;                        ///< Representing an invalid value
static const Year  INVALID_YEAR            = -1;                        ///< Representation of an invalid year
static const Date  INVALID_DATE            = -1;                        ///< Representation of an invalid date
static const Ticks INVALID_TICKS           = -1;                        ///< Representation of an invalid number of ticks

/*
 * ORIGINAL_BASE_YEAR, ORIGINAL_MAX_YEAR and DAYS_TILL_ORIGINAL_BASE_YEAR are
 * primarily used for loading newgrf and savegame data and returning some
 * newgrf (callback) functions that were in the original (TTD) inherited
 * format, where '_date == 0' meant that it was 1920-01-01.
 */

/** The minimum starting year/base year of the original TTD */
static const Year ORIGINAL_BASE_YEAR = 1920;
/** The original ending year */
static const Year ORIGINAL_END_YEAR  = 2051;
/** The maximum year of the original TTD */
static const Year ORIGINAL_MAX_YEAR  = 2090;

/**
 * Calculate the number of leap years till a given year.
 *
 * Each passed leap year adds one day to the 'day count'.
 *
 * A special case for the year 0 as no year has been passed,
 * but '(year - 1) / 4' does not yield '-1' to counteract the
 * '+1' at the end of the formula as divisions round to zero.
 *
 * @param year the year to get the leap years till.
 * @return the number of leap years.
 */
#define LEAP_YEARS_TILL(year) ((year) == 0 ? 0 : ((year) - 1) / 4 - ((year) - 1) / 100 + ((year) - 1) / 400 + 1)

/**
 * Calculate the date of the first day of a given year.
 * @param year the year to get the first day of.
 * @return the date.
 */
#define DAYS_TILL(year) (DAYS_IN_YEAR * (year) + LEAP_YEARS_TILL(year))

/**
 * The offset in days from the '_date == 0' till
 * 'ConvertYMDToDate(ORIGINAL_BASE_YEAR, 0, 1)'
 */
#define DAYS_TILL_ORIGINAL_BASE_YEAR DAYS_TILL(ORIGINAL_BASE_YEAR)

/** The absolute minimum & maximum years in OTTD */
static const Year MIN_YEAR = 0;

/** The default starting year */
static const Year DEF_START_YEAR = 1950;
/** The default scoring end year */
static const Year DEF_END_YEAR = ORIGINAL_END_YEAR - 1;

/**
 * MAX_YEAR, nicely rounded value of the number of years that can
 * be encoded in a single 32 bits date, about 2^31 / 366 years.
 */
static const Year MAX_YEAR  = 5000000;

/** The number of days till the last day */
#define MAX_DAY (DAYS_TILL(MAX_YEAR + 1) - 1)

/**
 * Data structure to convert between Date and triplet (year, month, and day).
 * @see ConvertDateToYMD(), ConvertYMDToDate()
 */
struct DateTime {
	Year   year;   ///< Year (0...)
	Month  month;  ///< Month (0..11)
	Day    day;    ///< Day (1..31)
	Hour   hour;   ///< Hour (00-24)
	Minute minute; ///< Minute (0-59)
};

#endif /* DATE_TYPE_H */
