/*
 * This file is part of OpenTTD.
 * OpenTTD is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, version 2.
 * OpenTTD is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details. You should have received a copy of the GNU General Public License along with OpenTTD. If not, see <http://www.gnu.org/licenses/>.
 */

/** @file loadunload.cpp Handling unloading and loading of cargo. */



#include "../stdafx.h"

#include "loadunload_static.h"
#include "loadunload_func.h"

#include "../cargomonitor.h"
#include "../company_func.h"
#include "../command_func.h"
#include "../economy_func.h"
#include "../industry.h"
#include "../newgrf_airporttiles.h"
#include "../newgrf_callbacks.h"
#include "../newgrf_industries.h"
#include "../newgrf_station.h"
#include "../town.h"
#include "../core/backup_type.hpp"
#include "../linkgraph/refresh.h"

#include "../engine_type.h"
#include "../vehicle_func.h"
#include "../vehicle_gui.h"
#include "../vehicle/aircraft.h"
#include "../vehicle/roadveh.h"
#include "../vehicle/ship.h"
#include "../vehicle/train.h"

#include "../debug.h"
#include "../safeguards.h"



/* begin static */

static void PrepareUnload         (Vehicle *front_v);
static void LoadUnloadVehicle     (Vehicle *front);
static void LoadUnload_A          (Vehicle *front, Station *st, LoadingData *loading_data,
								   CargoArray *consist_capleft, StationIDStack *next_station)
;
static uint GetLoadAmount         (Vehicle *vehicle);
static void HandleStationRefit    (Vehicle *vehicle, CargoArray &consist_capleft, Station *st,
								   StationIDStack next_station, CargoID new_cid)
;
static void ReserveConsist        (Station *station, Vehicle *u, CargoArray *consist_capleft,
								   StationIDStack *next_station)
;
static void UpdateLoadUnloadTicks (Vehicle *front, const Station *station, TimeTick ticks,
								   bool cargo_passengers)
;

/* end static */





template<class Taction>
bool IterateVehicleParts(Vehicle *vehicle, Taction action);





/**
 * Prepare everything to begin the loading when arriving at a station.
 * @pre IsTileType(this->tile, MP_STATION) || this->type == VEH_SHIP.
 */
void BeginLoading (
	Vehicle *vehicle
) {

//Debug (runtime, 3, "beginloading - enter {}", vehicle->unitnumber);
	assert(IsTileType(vehicle->tile, MP_STATION) || vehicle->type == VEH_SHIP);

	TimeTick travel_time = vehicle->current_order_time;
	if (vehicle->current_order.IsType(OT_GOTO_STATION)
		&& vehicle->current_order.dest == vehicle->last_station_visited
	) {
		if ( !(vehicle->current_order.GetUnloadType() & OUFB_NO_UNLOAD) ) {
			vehicle->vehstatus |= VS_UNLOADING;
		}
		if ( !(vehicle->current_order.GetLoadType() & OLFB_NO_LOAD) ) {
			vehicle->vehstatus |= VS_LOADING;
		}

		vehicle->current_order.SetNonStopType(ONSF_NO_STOP_AT_ANY_STATION);

		if (vehicle->vehstatus |= VS_UNLOADING | VS_LOADING) {
			vehicle->vehstatus &= ~VS_TRAVELLING;
			PrepareUnload(vehicle);
		} else {
			HandleWaiting(vehicle);
		}
	}

	if (
		vehicle->last_loading_station != INVALID_STATION
		&& vehicle->last_loading_station != vehicle->last_station_visited
		&& (
			( !(vehicle->current_order.GetLoadType() & OLFB_NO_LOAD) )
			|| ( !(vehicle->current_order.GetUnloadType() & OUFB_NO_UNLOAD) )
		)
	) {
		IncreaseStats (
						Station::Get (vehicle->last_loading_station),
						vehicle,
						vehicle->last_station_visited,
						travel_time
		);
	}

	SetWindowDirty(GetWindowClassForVehicleType(vehicle->type), vehicle->owner);
	SetWindowWidgetDirty(WC_VEHICLE_VIEW, vehicle->index, WID_VV_START_STOP);
	SetWindowDirty(WC_VEHICLE_DETAILS, vehicle->index);
	SetWindowDirty(WC_STATION_VIEW, vehicle->last_station_visited);

	Station::Get(vehicle->last_station_visited)->MarkTilesDirty(true);
	vehicle->cur_speed = 0;
	vehicle->MarkDirty();

//Debug (runtime, 3, "beginloading - exit");
}





/**
 * Prepare the vehicle to be unloaded.
 * @param front_v the vehicle to be unloaded
 */

static void PrepareUnload (
	Vehicle *front_v
) {

	Station *curr_station = Station::Get(front_v->last_station_visited);
	curr_station->loading_vehicles.push_back(front_v);
	front_v->vehstatus |= VS_STATION_LISTED;

	/* Start unloading at the first possible moment */
	front_v->load_unload_ticks = 1;

	/* One CargoPayment per vehicle and the vehicle limit equals the
	 * limit in number of CargoPayments. Can't go wrong. */
//	static_assert(CargoPaymentPool::MAX_SIZE == VehiclePool::MAX_SIZE);
	front_v->cargo_payment = new CargoPayment(front_v);

	StationIDStack next_station = front_v->GetNextStoppingStation();

	if (front_v->orders.list == nullptr
		|| ( !(front_v->current_order.GetUnloadType() & OUFB_NO_UNLOAD) )
	) {
		Station *station = Station::Get(front_v->last_station_visited);

		for (Vehicle *vehicle = front_v; vehicle != nullptr; vehicle = vehicle->Next()) {
			const GoodsEntry *ge = &station->goods [vehicle->cargo_type];

			if (vehicle->cargo_cap > 0 && vehicle->cargo.TotalCount() > 0) {
				vehicle->cargo.Stage(
					HasBit (ge->status, GoodsEntry::GES_ACCEPTANCE),
					front_v->last_station_visited, next_station,
					front_v->current_order.GetUnloadType(), ge,
					front_v->cargo_payment
				);

				if (vehicle->cargo.UnloadCount() > 0) vehicle->vehicle_flags |= VF_CARGO_UNLOADING;
			}
		}
	}

//Debug (runtime, 3, "prepareunload - exit");
}





/**
 * Load/unload the vehicles in this station according to the order
 * they entered.
 * @param st the station to do the loading/unloading for
 */

void LoadUnloadStation (
	Station *station
) {

	std::list<Vehicle *>::iterator iter;
	for (iter = station->loading_vehicles.begin(); iter != station->loading_vehicles.end(); ++iter) {
		Vehicle *vehicle = *iter;
		if (--vehicle->load_unload_ticks == 0
			&& (!(vehicle->vehstatus & (VS_STOPPED | VS_CRASHED | VS_WAITING)))
			&& (vehicle->vehstatus & (VS_UNLOADING | VS_LOADING))
		) {
			LoadUnloadVehicle(vehicle);
		}

	if (station->loading_vehicles.empty()) break;
	}

	/* Call the production machinery of industries */
	for (Industry *iid : _cargo_delivery_destinations) {
		TriggerIndustryProduction(iid);
	}
	_cargo_delivery_destinations.clear();

//Debug (runtime, 3, "loadunload station - exit");
}





/**
 * Loads/unload the vehicle if possible.
 * @param front the vehicle to be (un)loaded
 */
static void LoadUnloadVehicle(Vehicle *front)
{
	StationID last_visited = front->last_station_visited;
	Station *station = Station::Get(last_visited);

	if (front->type == VEH_TRAIN
		&& (!IsTileType(front->tile, MP_STATION)
			|| GetStationIndex(front->tile) != station->index)
	) {
		/* The train reversed in the station. Take the "easy" way
		 * out and let the train just leave as it always did. */
		front->vehstatus &= ~VS_UNLOADING;
		front->vehstatus &= ~VS_LOADING;
		HandleWaiting(front);
		front->load_unload_ticks = 1;

		return;
	}

	LoadingData loading_data = {
		0, 0,
		false, false,
		true,
		false, false,
		0, 0, 0, 0,
		front->cargo_payment
	};

	StationIDStack next_station = front->GetNextStoppingStation();
	bool use_autorefit = front->current_order.IsRefit() && front->current_order.GetRefitCargo() == CT_AUTO_REFIT;
	CargoArray consist_capleft;

	if (_settings_game.order.improved_load && use_autorefit
			? front->cargo_payment == nullptr
			: (front->current_order.GetLoadType() & OLFB_FULL_LOAD) != 0
	) {
		ReserveConsist(station, front,
						(use_autorefit && front->load_unload_ticks != 0)
							? &consist_capleft
							: nullptr,
						&next_station);
	}

	front->cur_speed = 0;

	LoadUnload_A(front, station, &loading_data, &consist_capleft, &next_station);

	if (front->type == VEH_TRAIN) {
		if (loading_data.anything_loaded || loading_data.anything_unloaded) {
			TriggerStationRandomisation(station, front->tile, SRT_TRAIN_LOADS);
			TriggerStationAnimation(station, front->tile, SAT_TRAIN_LOADS);
		}
	}

	/* Only set completely_emptied, if we just unloaded all remaining cargo */
	loading_data.completely_emptied &= loading_data.anything_unloaded;

	if (!loading_data.anything_unloaded) delete loading_data.payment;

	front->vehicle_flags &= ~VF_STOP_LOADING;
	bool cargo_passengers = IsCargoInClass(front->cargo_type, CC_PASSENGERS);
	if (loading_data.anything_loaded || loading_data.anything_unloaded) {
		if (_settings_game.order.gradual_loading) {
			loading_data.new_load_unload_ticks = gradual_loading_wait_time[front->type];
		}
		/* We loaded less cargo than possible for all cargo types and it's not full
		 * load and we're not supposed to wait any longer: stop loading. */
		if (!loading_data.anything_unloaded && loading_data.full_load_amount == 0
			&& loading_data.reservation_left == 0 && !(front->current_order.GetLoadType() & OLFB_FULL_LOAD)
			&& front->current_order_time >= std::max(front->current_order.wait_time - front->lateness_counter, 0LL)
		) {
			front->vehicle_flags |= VF_STOP_LOADING;
		}

		UpdateLoadUnloadTicks(front, station, std::max(1LL, loading_data.new_load_unload_ticks), cargo_passengers);
	} else {
		UpdateLoadUnloadTicks(front, station, 20, cargo_passengers); // We need the ticks for link refreshing.
		bool finished_loading = true;
		if (front->current_order.GetLoadType() & OLFB_FULL_LOAD) {
			if (front->current_order.GetLoadType() == OLF_FULL_LOAD_ANY) {
				/* if the aircraft carries passengers and is NOT full, then
				 * continue loading, no matter how much mail is in */
				if ((front->type == VEH_AIRCRAFT && cargo_passengers && front->cargo_cap > front->cargo.StoredCount()) ||
						(loading_data.cargo_not_full != 0 && (loading_data.cargo_full & ~loading_data.cargo_not_full) == 0)) {
					// There are still non-full cargoes
					finished_loading = false;
				}
			} else if (loading_data.cargo_not_full != 0) {
				finished_loading = false;
			}

			/* Refresh next hop stats if we're full loading to make the links
			 * known to the distribution algorithm and allow cargo to be sent
			 * along them. Otherwise the vehicle could wait for cargo
			 * indefinitely if it hasn't visited the other links yet, or if the
			 * links die while it's loading. */
			if (!finished_loading) LinkRefresher::Run(front, true, true);
		}

		if (finished_loading) {
			front->vehstatus &= ~VS_UNLOADING;
			front->vehstatus &= ~VS_LOADING;
//			front->vehstatus |=  VS_WAITING;
			HandleWaiting(front);
		}
	}

	/* Calculate the loading indicator fill percent and display
	 * In the Game Menu do not display indicators
	 * If _settings_client.gui.loading_indicators == 2, show indicators (bool can be promoted to int as 0 or 1 - results in 2 > 0,1 )
	 * if _settings_client.gui.loading_indicators == 1, _local_company must be the owner or must be a spectator to show ind., so 1 > 0
	 * if _settings_client.gui.loading_indicators == 0, do not display indicators ... 0 is never greater than anything
	 * Hide display when thee's no loading and unloading order.
	 */
	if (!(front->current_order.GetLoadType() & OLFB_NO_LOAD)) {
		if (_game_mode != GM_MENU && (_settings_client.gui.loading_indicators > (uint)(front->owner != _local_company && _local_company != COMPANY_SPECTATOR))) {
			StringID percent_up_down = STR_NULL;
			int percent = CalcPercentVehicleFilled(front, &percent_up_down);
			if (front->fill_percent_te_id == INVALID_TE_ID) {
				front->fill_percent_te_id = ShowFillingPercent(front->x_pos, front->y_pos, front->z_pos + 20, percent, percent_up_down);
			} else {
				UpdateFillingPercent(front->fill_percent_te_id, percent, percent_up_down);
			}
		}
	}

	if (loading_data.completely_emptied) {
		/* Make sure the vehicle is marked dirty, since we need to update the NewGRF
		 * properties such as weight, power and TE whenever the trigger runs. */
		loading_data.dirty_vehicle = true;
		TriggerVehicle(front, VEHICLE_TRIGGER_EMPTY);
	}

	if (loading_data.dirty_vehicle) {
		SetWindowDirty(GetWindowClassForVehicleType(front->type), front->owner);
		SetWindowDirty(WC_VEHICLE_DETAILS, front->index);
		front->MarkDirty();
	}
	if (loading_data.dirty_station) {
		station->MarkTilesDirty(true);
		SetWindowDirty(WC_STATION_VIEW, last_visited);
		InvalidateWindowData(WC_STATION_LIST, last_visited);
	}
//Debug (runtime, 3, "loadunload vehicle - exit");
}





static void LoadUnload_A(Vehicle *front, Station *station, LoadingData *loading_data, CargoArray *consist_capleft, StationIDStack *next_station)
{
	for (Vehicle *v = front; v != nullptr; v = v->Next()) {
		if (v == front || !v->Previous()->HasArticulatedPart()) loading_data->artic_part = 0;
		if (v->cargo_cap == 0) continue;
		++loading_data->artic_part;

		GoodsEntry *ge = &station->goods[v->cargo_type];

		if ((v->vehicle_flags & VF_CARGO_UNLOADING) && (front->current_order.GetUnloadType() & OUFB_NO_UNLOAD) == 0) {
			uint cargo_count = v->cargo.UnloadCount();
			uint amount_unloaded = _settings_game.order.gradual_loading ? std::min(cargo_count, GetLoadAmount(v)) : cargo_count;
			bool remaining = false; // Are there cargo entities in this vehicle that can still be unloaded here?

			assert(payment != nullptr);
			loading_data->payment->SetCargo(v->cargo_type);

			if (!HasBit(ge->status, GoodsEntry::GES_ACCEPTANCE) && v->cargo.ActionCount(VehicleCargoList::MTA_DELIVER) > 0) {
				/* The station does not accept our goods anymore. */
				if (front->current_order.GetUnloadType() & (OUFB_TRANSFER | OUFB_UNLOAD)) {
					/* Transfer instead of delivering. */
					v->cargo.Reassign<VehicleCargoList::MTA_DELIVER, VehicleCargoList::MTA_TRANSFER>(
							v->cargo.ActionCount(VehicleCargoList::MTA_DELIVER), INVALID_STATION);
				} else {
					uint new_remaining = v->cargo.RemainingCount() + v->cargo.ActionCount(VehicleCargoList::MTA_DELIVER);
					if (v->cargo_cap < new_remaining) {
						/* Return some of the reserved cargo to not overload the vehicle. */
						v->cargo.Return(new_remaining - v->cargo_cap, &ge->cargo, INVALID_STATION);
					}

					/* Keep instead of delivering. This may lead to no cargo being unloaded, so ...*/
					v->cargo.Reassign<VehicleCargoList::MTA_DELIVER, VehicleCargoList::MTA_KEEP>(
							v->cargo.ActionCount(VehicleCargoList::MTA_DELIVER));

					/* ... say we unloaded something, otherwise we'll think we didn't unload
					 * something and we didn't load something, so we must be finished
					 * at this station. Setting the unloaded means that we will get a
					 * retry for loading in the next cycle. */
					loading_data->anything_unloaded = true;
				}
			}

			if (v->cargo.ActionCount(VehicleCargoList::MTA_TRANSFER) > 0) {
				/* Mark the station dirty if we transfer, but not if we only deliver. */
				loading_data->dirty_station = true;

				if (!ge->HasRating()) {
					/* Upon transferring cargo, make sure the station has a rating. Fake a pickup for the
					 * first unload to prevent the cargo from quickly decaying after the initial drop. */
					ge->time_since_pickup = 0;
					SetBit(ge->status, GoodsEntry::GES_RATING);
				}
			}

			amount_unloaded = v->cargo.Unload(amount_unloaded, &ge->cargo, loading_data->payment);
			remaining = v->cargo.UnloadCount() > 0;
			if (amount_unloaded > 0) {
				loading_data->dirty_vehicle = true;
				loading_data->anything_unloaded = true;
				loading_data->new_load_unload_ticks += amount_unloaded;

				/* Deliver goods to the station */
				station->time_since_unload = 0;
			}

			if (_settings_game.order.gradual_loading && remaining) {
				loading_data->completely_emptied = false;
			} else {
				/* We have finished unloading (cargo count == 0) */
				v->vehicle_flags &= ~VF_CARGO_UNLOADING;
			}

			continue;
		}

		/* Do not pick up goods when we have no-load set or loading is stopped. */
		if ((front->current_order.GetLoadType() & OLFB_NO_LOAD) || (front->vehicle_flags & VF_STOP_LOADING)) continue;

		/* This order has a refit, if this is the first vehicle part carrying cargo and the whole vehicle is empty, try refitting. */
		if (front->current_order.IsRefit() && loading_data->artic_part == 1) {
			HandleStationRefit(v, *consist_capleft, station, *next_station, front->current_order.GetRefitCargo());
			ge = &station->goods[v->cargo_type];
		}

		/* As we're loading here the following link can carry the full capacity of the vehicle. */
		v->refit_cap = v->cargo_cap;

		/* update stats */
		int t;
		switch (front->type) {
			case VEH_TRAIN:
			case VEH_SHIP:
				t = front->vcache.cached_max_speed;
				break;

			case VEH_ROAD:
				t = front->vcache.cached_max_speed / 2;
				break;

			case VEH_AIRCRAFT:
				t = Aircraft::From(front)->GetSpeedOldUnits(); // Convert to old units.
				break;

			default: NOT_REACHED();
		}

		/* if last speed is 0, we treat that as if no vehicle has ever visited the station. */
		ge->last_speed = std::min(t, 255);
		ge->last_age = std::min(_cur_year - front->build_year, 255);

		assert(v->cargo_cap >= v->cargo.StoredCount());
		/* Capacity available for loading more cargo. */
		uint cap_left = v->cargo_cap - v->cargo.StoredCount();
		if (cap_left > 0) {
			/* If vehicle can load cargo, reset time_since_pickup. */
			ge->time_since_pickup = 0;

			/* If there's goods waiting at the station, and the vehicle
			 * has capacity for it, load it on the vehicle. */
			if ((v->cargo.ActionCount(VehicleCargoList::MTA_LOAD) > 0
					|| ge->cargo.AvailableCount() > 0)
				&& MayLoadUnderExclusiveRights(station, v)
			) {
				if (v->cargo.StoredCount() == 0) {
					TriggerVehicle(v, VEHICLE_TRIGGER_NEW_CARGO);
				}
				if (_settings_game.order.gradual_loading) cap_left = std::min(cap_left, GetLoadAmount(v));

				uint loaded = ge->cargo.Load(cap_left, &v->cargo, station->xy, *next_station);
				if (v->cargo.ActionCount(VehicleCargoList::MTA_LOAD) > 0) {
					/* Remember if there are reservations left so that we don't stop
					 * loading before they're loaded. */
					SetBit(loading_data->reservation_left, v->cargo_type);
				}

				/* Store whether the maximum possible load amount was loaded or not.*/
				if (loaded == cap_left) {
					SetBit(loading_data->full_load_amount, v->cargo_type);
				} else {
					ClrBit(loading_data->full_load_amount, v->cargo_type);
				}

				/* TODO: Regarding this, when we do gradual loading, we
				 * should first unload all vehicles and then start
				 * loading them. Since this will cause
				 * VEHICLE_TRIGGER_EMPTY to be called at the time when
				 * the whole vehicle chain is really totally empty, the
				 * completely_emptied assignment can then be safely
				 * removed; that's how TTDPatch behaves too. --pasky */
				if (loaded > 0) {
					loading_data->completely_emptied = false;
					loading_data->anything_loaded = true;

					station->time_since_load = 0;
					station->last_vehicle_type = v->type;

					if (ge->cargo.TotalCount() == 0) {
						TriggerStationRandomisation(station, station->xy, SRT_CARGO_TAKEN, v->cargo_type);
						TriggerStationAnimation(station, station->xy, SAT_CARGO_TAKEN, v->cargo_type);
						AirportAnimationTrigger(station, AAT_STATION_CARGO_TAKEN, v->cargo_type);
					}

					loading_data->new_load_unload_ticks += loaded;

					loading_data->dirty_vehicle = loading_data->dirty_station = true;
				}
			}
		}

		if (v->cargo.StoredCount() >= v->cargo_cap) {
			SetBit(loading_data->cargo_full, v->cargo_type);
		} else {
			SetBit(loading_data->cargo_not_full, v->cargo_type);
		}
	}
//Debug (runtime, 3, "loadunload sub A - exit");
}





/**
 * Reserves cargo if the full load order and improved_load is set or if the
 * current order allows autorefit.
 * @param st Station where the consist is loading at the moment.
 * @param v Front of the loading vehicle consist.
 * @param consist_capleft If given, save free capacities after reserving there.
 * @param next_station Station(s) the vehicle will stop at next.
 */
static void ReserveConsist(Station *st, Vehicle *v, CargoArray *consist_capleft, StationIDStack *next_station)
{
	/* If there is a cargo payment not all vehicles of the consist have tried to do the refit.
	 * In that case, only reserve if it's a fixed refit and the equivalent of "articulated chain"
	 * a vehicle belongs to already has the right cargo. */
	bool must_reserve = !v->current_order.IsRefit() || v->cargo_payment == nullptr;
	for (Vehicle *u = v; u != nullptr; u = u->Next()) {
		assert(u->cargo_cap >= u->cargo.RemainingCount());

		/* Exclude various ways in which the vehicle might not be the head of an equivalent of
		 * "articulated chain". Also don't do the reservation if the vehicle is going to refit
		 * to a different cargo and hasn't tried to do so, yet. */
		if (!u->IsArticulatedPart()
			&& (u->type != VEH_TRAIN || !Train::From(u)->IsRearDualheaded())
			&& (u->type != VEH_AIRCRAFT || Aircraft::From(u)->IsNormalAircraft())
			&& (must_reserve || u->current_order.GetRefitCargo() == u->cargo_type)
		) {
			IterateVehicleParts(u, ReserveCargoAction(st, next_station));
		}
		if (consist_capleft == nullptr || u->cargo_cap == 0) continue;
		(*consist_capleft)[u->cargo_type] += u->cargo_cap - u->cargo.RemainingCount();
	}
}





/**
 * Gets the amount of cargo the given vehicle can load in the current tick.
 * This is only about loading speed. The free capacity is ignored.
 * @param v Vehicle to be queried.
 * @return Amount of cargo the vehicle can load at once.
 */
static uint GetLoadAmount(Vehicle *v)
{
	const Engine *e = v->GetEngine();
	uint load_amount = e->info.load_amount;

	/* The default loadamount for mail is 1/4 of the load amount for passengers */
	bool air_mail = v->type == VEH_AIRCRAFT && !Aircraft::From(v)->IsNormalAircraft();
	if (air_mail) load_amount = CeilDiv(load_amount, 4);

	if (_settings_game.order.gradual_loading) {
		uint16 cb_load_amount = CALLBACK_FAILED;
		if (e->GetGRF() != nullptr && e->GetGRF()->grf_version >= 8) {
			/* Use callback 36 */
			cb_load_amount = GetVehicleProperty(v, PROP_VEHICLE_LOAD_AMOUNT, CALLBACK_FAILED);
		} else if (HasBit(e->info.callback_mask, CBM_VEHICLE_LOAD_AMOUNT)) {
			/* Use callback 12 */
			cb_load_amount = GetVehicleCallback(CBID_VEHICLE_LOAD_AMOUNT, 0, 0, v->engine_type, v);
		}
		if (cb_load_amount != CALLBACK_FAILED) {
			if (e->GetGRF()->grf_version < 8) cb_load_amount = GB(cb_load_amount, 0, 8);
			if (cb_load_amount >= 0x100) {
				ErrorUnknownCallbackResult(e->GetGRFID(), CBID_VEHICLE_LOAD_AMOUNT, cb_load_amount);
			} else if (cb_load_amount != 0) {
				load_amount = cb_load_amount;
			}
		}
	}

	/* Scale load amount the same as capacity */
	if (HasBit(e->info.misc_flags, EF_NO_DEFAULT_CARGO_MULTIPLIER) && !air_mail) load_amount = CeilDiv(load_amount * CargoSpec::Get(v->cargo_type)->multiplier, 0x100);

	/* Zero load amount breaks a lot of things. */
	return std::max(1u, load_amount);
}





/**
 * Iterate the articulated parts of a vehicle, also considering the special cases of "normal"
 * aircraft and double headed trains. Apply an action to each vehicle and immediately return false
 * if that action does so. Otherwise return true.
 * @tparam Taction Class of action to be applied. Must implement bool operator()([const] Vehicle *).
 * @param v First articulated part.
 * @param action Instance of Taction.
 * @return false if any of the action invocations returned false, true otherwise.
 */
template<class Taction>
bool IterateVehicleParts(Vehicle *v, Taction action)
{
	for (Vehicle *w = v; w != nullptr;
			w = w->HasArticulatedPart() ? w->GetNextArticulatedPart() : nullptr) {
		if (!action(w)) return false;
		if (w->type == VEH_TRAIN) {
			Train *train = Train::From(w);
			if (train->IsMultiheaded() && !action(train->other_multiheaded_part)) return false;
		}
	}
	if (v->type == VEH_AIRCRAFT && Aircraft::From(v)->IsNormalAircraft()) return action(v->Next());
	return true;
}





/**
 * Refit a vehicle in a station.
 * @param v Vehicle to be refitted.
 * @param consist_capleft Added cargo capacities in the consist.
 * @param st Station the vehicle is loading at.
 * @param next_station Possible next stations the vehicle can travel to.
 * @param new_cid Target cargo for refit.
 */
static void HandleStationRefit(Vehicle *v, CargoArray &consist_capleft, Station *st, StationIDStack next_station, CargoID new_cid)
{
	Vehicle *v_start = v->GetFirstEnginePart();
	if (!IterateVehicleParts(v_start, IsEmptyAction())) return;

	Backup<CompanyID> cur_company(_current_company, v->owner, FILE_LINE);

	CargoTypes refit_mask = v->GetEngine()->info.refit_mask;

	/* Remove old capacity from consist capacity and collect refit mask. */
	IterateVehicleParts(v_start, PrepareRefitAction(consist_capleft, refit_mask));

	bool is_auto_refit = new_cid == CT_AUTO_REFIT;
	if (is_auto_refit) {
		/* Get a refittable cargo type with waiting cargo for next_station or INVALID_STATION. */
		new_cid = v_start->cargo_type;
		for (CargoID cid : SetCargoBitIterator(refit_mask)) {
			if (st->goods[cid].cargo.HasCargoFor(next_station)) {
				/* Try to find out if auto-refitting would succeed. In case the refit is allowed,
				 * the returned refit capacity will be greater than zero. */
				CommandContainer cmdcont;
					cmdcont.tile = v_start->tile;
					cmdcont.p1 = v_start->index;
					cmdcont.p2 = cid | 1U << 24 | 0xFF << 8 | 1U << 16;
					cmdcont.dc_flags = DC_QUERY_COST;
					cmdcont.cmd = GetCmdRefitVeh(v_start);
				ProcessCmd(cmdcont); // Auto-refit and only this vehicle including artic parts.
				/* Try to balance different loadable cargoes between parts of the consist, so that
				 * all of them can be loaded. Avoid a situation where all vehicles suddenly switch
				 * to the first loadable cargo for which there is only one packet. If the capacities
				 * are equal refit to the cargo of which most is available. This is important for
				 * consists of only a single vehicle as those will generally have a consist_capleft
				 * of 0 for all cargoes. */
				if (_returned_refit_capacity > 0 && (consist_capleft[cid] < consist_capleft[new_cid] ||
						(consist_capleft[cid] == consist_capleft[new_cid] &&
						st->goods[cid].cargo.AvailableCount() > st->goods[new_cid].cargo.AvailableCount()))) {
					new_cid = cid;
				}
			}
		}
	}

	/* Refit if given a valid cargo. */
	if (new_cid < NUM_CARGO && new_cid != v_start->cargo_type) {
		/* INVALID_STATION because in the DT_MANUAL case that's correct and in the DT_(A)SYMMETRIC
		 * cases the next hop of the vehicle doesn't really tell us anything if the cargo had been
		 * "via any station" before reserving. We rather produce some more "any station" cargo than
		 * misrouting it. */
		IterateVehicleParts(v_start, ReturnCargoAction(st, INVALID_STATION));
		CommandContainer cmdcont;
			cmdcont.tile = v_start->tile;
			cmdcont.p1 =  v_start->index;
			cmdcont.p2 = new_cid | 1U << 24 | 0xFF << 8 | 1U << 16;
			cmdcont.dc_flags = DC_EXEC;
			cmdcont.cmd = GetCmdRefitVeh(v_start);
		CommandCost cost = ProcessCmd(cmdcont);
		if (cost.Succeeded()) v->First()->profit_this_year -= cost.GetCost() << 8;
	}

	/* Add new capacity to consist capacity and reserve cargo */
	IterateVehicleParts(v_start, FinalizeRefitAction(consist_capleft, st, next_station,
			is_auto_refit || (v->First()->current_order.GetLoadType() & OLFB_FULL_LOAD) != 0));

	cur_company.Restore();
}





/**
 * Test whether a vehicle can load cargo at a station even if exclusive transport rights are present.
 * @param st Station with cargo waiting to be loaded.
 * @param v Vehicle loading the cargo.
 * @return true when a vehicle can load the cargo.
 */
static bool MayLoadUnderExclusiveRights(const Station *st, const Vehicle *v)
{
	return st->owner != OWNER_NONE || st->town->exclusive_counter == 0 || st->town->exclusivity == v->owner;
}





/**
 * Update the vehicle's load_unload_ticks, the time it will wait until it tries to load or unload
 * again. Adjust for length of trains and set it at least to 1.
 * @param front The vehicle to be updated.
 * @param st The station the vehicle is loading at.
 * @param ticks The time it would normally wait, based on cargo loaded and unloaded.
 */
static void UpdateLoadUnloadTicks(Vehicle *front, const Station *st, TimeTick ticks, bool cargo_passengers)
{
	switch (front->type) {
		case VEH_TRAIN: {
			/* Time increases by length of train as it won't be loaded/unloaded by its full length in reality. */
			ticks += (front->GetGroundVehicleCache()->cached_total_length * ticks) / 10;
			break;
		}

		case VEH_ROAD: {
			ticks += (_scale * (_scale / 3));
			break;
		}

		case VEH_AIRCRAFT: {
			/* Fast travel but long boarding time. */
			ticks += (_scale * _scale * (_scale / 3));
			break;
		}

		default: {
			ticks += (_scale * _scale);
			break;
 		}
 	}

	if (cargo_passengers) ticks /= _scale;
	/* Semi-randomise loading time. */
	if (_tick_loop % 3) ticks += 33;
	/* Always wait at least 1, otherwise we'll wait 'infinitively' long. */
	front->load_unload_ticks = std::max(1LL, (ticks *= _scale));
}





/**
 * Transfer goods from station to industry.
 * All cargo is delivered to the nearest (Manhattan) industry to the station sign, which is inside the acceptance rectangle and actually accepts the cargo.
 * @param st The station that accepted the cargo
 * @param cargo_type Type of cargo delivered
 * @param num_pieces Amount of cargo delivered
 * @param source The source of the cargo
 * @param company The company delivering the cargo
 * @return actually accepted pieces of cargo
 */
uint DeliverGoodsToIndustry(const Station *st, CargoID cargo_type, uint num_pieces, IndustryID source, CompanyID company)
{
	/* Find the nearest industrytile to the station sign inside the catchment area, whose industry accepts the cargo.
	 * This fails in three cases:
	 *  1) The station accepts the cargo because there are enough houses around it accepting the cargo.
	 *  2) The industries in the catchment area temporarily reject the cargo, and the daily station loop has not yet updated station acceptance.
	 *  3) The results of callbacks CBID_INDUSTRY_REFUSE_CARGO and CBID_INDTILE_CARGO_ACCEPTANCE are inconsistent. (documented behaviour)
	 */

	uint accepted = 0;

	for (Industry *ind : st->industries_near) {
		if (num_pieces == 0) break;

		if (ind->index == source) continue;

		uint cargo_index;
		for (cargo_index = 0; cargo_index < lengthof(ind->accepts_cargo); cargo_index++) {
			if (cargo_type == ind->accepts_cargo[cargo_index]) break;
		}
		/* Check if matching cargo has been found */
		if (cargo_index >= lengthof(ind->accepts_cargo)) continue;

		/* Check if industry temporarily refuses acceptance */
		if (IndustryTemporarilyRefusesCargo(ind, cargo_type)) continue;

		if (ind->exclusive_supplier != INVALID_OWNER && ind->exclusive_supplier != st->owner) continue;

		/* Insert the industry into _cargo_delivery_destinations, if not yet contained */
		include(_cargo_delivery_destinations, ind);

		uint amount = std::min(num_pieces, 0xFFFFu - ind->incoming_cargo_waiting[cargo_index]);
		ind->incoming_cargo_waiting[cargo_index] += amount;
		ind->last_cargo_accepted_at[cargo_index] = _date;
		num_pieces -= amount;
		accepted += amount;

		/* Update the cargo monitor. */
		AddCargoDelivery(cargo_type, company, amount, ST_INDUSTRY, source, st, ind->index);
	}

	return accepted;
}





/**
 * Return all reserved cargo packets to the station and reset all packets
 * staged for transfer.
 * @param st the station where the reserved packets should go.
 */
void Vehicle::CancelReservation(StationID next, Station *st)
{
	for (Vehicle *v = this; v != nullptr; v = v->next) {
		VehicleCargoList &cargo = v->cargo;
		if (cargo.ActionCount(VehicleCargoList::MTA_LOAD) > 0) {
			Debug(misc, 1, "cancelling cargo reservation");
			cargo.Return(UINT_MAX, &st->goods[v->cargo_type].cargo, next);
			cargo.SetTransferLoadPlace(st->xy);
		}
		cargo.KeepAll();
	}
}





/**
 * Get a map of cargoes and free capacities in the consist.
 * @param capacities Map to be filled with cargoes and capacities.
 */
void Vehicle::GetConsistFreeCapacities(SmallMap<CargoID, uint> &capacities) const
{
	for (const Vehicle *v = this; v != nullptr; v = v->Next()) {
		if (v->cargo_cap == 0) continue;
		std::pair<CargoID, uint> *pair = capacities.Find(v->cargo_type);
		if (pair == capacities.End()) {
			capacities.push_back({v->cargo_type, v->cargo_cap - v->cargo.StoredCount()});
		} else {
			pair->second += v->cargo_cap - v->cargo.StoredCount();
		}
	}
}





uint Vehicle::GetConsistTotalCapacity() const
{
	uint result = 0;
	for (const Vehicle *v = this; v != nullptr; v = v->Next()) {
		result += v->cargo_cap;
	}
	return result;
}
