/*
 * This file is part of OpenTTD.
 * OpenTTD is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, version 2.
 * OpenTTD is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details. You should have received a copy of the GNU General Public License along with OpenTTD. If not, see <http://www.gnu.org/licenses/>.
 */

/** @file order_base.h Base class for orders. */



#ifndef ORDER_BASE_H
#define ORDER_BASE_H



#include "order_type.h"
#include "core/pool_type.hpp"
#include "core/bitmath_func.hpp"
#include "cargo_type.h"
#include "depot_type.h"
#include "station_type.h"
#include "vehicle_type.h"
#include "date_type.h"
#include "saveload/saveload.h"



typedef Pool<Order, OrderID, 256, 0xFF0000> OrderPool;
typedef Pool<OrderList, OrderListID, 128, 64000> OrderListPool;
extern OrderPool _order_pool;
extern OrderListPool _orderlist_pool;



/* If you change this, keep in mind that it is saved on 3 places:
 * - Load_ORDR, all the global orders
 * - Vehicle -> current_order
 * - REF_ORDER (all REFs are currently limited to 16 bits!!)
 */
struct Order : OrderPool::PoolItem<&_order_pool> {
//private:
	friend struct VEHSChunkHandler;                             ///< Loading of ancient vehicles.
	friend SaveLoadTable GetOrderDescription();                 ///< Saving and loading of orders.
	/* So we can use private/protected variables in the saveload code */
	friend class SlVehicleCommon;
	friend class SlVehicleDisaster;

	uint8         type;                ///< The type of order + non-stop flags
	uint8         flags;               ///< Load/unload types, depot order/action types.
	int64         order_flags;         ///< Flags specific to the order itself.
	DestinationID dest;                ///< The destination of the order.

	CargoID       refit_cargo;         ///< Refit CargoID

	TimeTick      timed_arrival;       ///< The time of actual arrival at the destination.
	TimeTick      timed_departure;     ///< The time of actual departure.
	TimeTick      scheduled_arrival;   ///< The time of scheduled arrival.
	TimeTick      scheduled_departure; ///< The time of scheduled departure.
	TimeTick      wait_time;           ///< How long in ticks to wait at the destination.
	TimeTick      travel_time;         ///< How long in ticks the journey to this destination should take.
	uint16        max_speed;           ///< How fast the vehicle may go on the way to the destination.

//public:
	Order         *next;               ///< Pointer to next order. If nullptr, end of list

	Order() : flags(0), refit_cargo(CT_NO_REFIT),
			  timed_arrival (0), timed_departure(0), scheduled_arrival(0), scheduled_departure(0),
			  wait_time(0), max_speed(UINT16_MAX)
	{}
	~Order();

	Order(uint32 packed);

	bool IsType(OrderType type) const;
	OrderType GetType() const;

	void Free();

	void MakeGoToStation(StationID destination);
	void MakeGoToDepot(DepotID destination, OrderDepotTypeFlags order, OrderNonStopFlags non_stop_type = ONSF_NO_STOP_AT_INTERMEDIATE_STATIONS, OrderDepotActionFlags action = ODATF_SERVICE_ONLY, CargoID cargo = CT_NO_REFIT);
	void MakeGoToWaypoint(StationID destination);
	void MakeLoading(bool ordered);
	void MakeLeaveStation();
	void MakeDummy();
	void MakeConditional(VehicleOrderID order);
	void MakeWaiting();

	bool IsGotoOrder() const;
	DestinationID GetDestination() const;
	void SetDestination(DestinationID destination);
	bool IsRefit() const;
	bool IsAutoRefit() const;

	CargoID GetRefitCargo() const;

	void SetRefit(CargoID cargo);
	OrderLoadFlags GetLoadType() const;
	OrderUnloadFlags GetUnloadType() const;
	OrderNonStopFlags GetNonStopType() const;
	OrderStopLocation GetStopLocation() const;
	OrderDepotTypeFlags GetDepotOrderType() const;
	OrderDepotActionFlags GetDepotActionType() const;
	OrderWaypointFlags GetWaypointFlags() const;
	OrderConditionVariable GetConditionVariable() const;
	OrderConditionComparator GetConditionComparator() const;
	VehicleOrderID GetConditionSkipToOrder() const;
	uint16 GetConditionValue() const;

	void SetLoadType(OrderLoadFlags load_type);
	void SetUnloadType(OrderUnloadFlags unload_type);
	void SetNonStopType(OrderNonStopFlags non_stop_type);
	void SetStopLocation(OrderStopLocation stop_location);
	void SetDepotOrderType(OrderDepotTypeFlags depot_order_type);
	void SetDepotActionType(OrderDepotActionFlags depot_service_type);
	void SetWaypointFlags(OrderWaypointFlags waypoint_flags);
	void SetConditionVariable(OrderConditionVariable condition_variable);
	void SetConditionComparator(OrderConditionComparator condition_comparator);
	void SetConditionSkipToOrder(VehicleOrderID order_id);
	void SetConditionValue(uint16 value);

//	bool IsWaitTimetabled() const;
//	bool IsTravelTimetabled() const;

//	void SetWaitTimetabled(TimeTick time);
//	void SetTravelTimetabled(bool timetabled);

	bool CanLoadOrUnload() const;
	bool CanLeaveWithCargo(bool has_cargo) const;

//	bool IsCompletelyTimetabled() const;

	void AssignOrder(const Order &other);
	bool Equals(const Order &other) const;

	uint32 Pack() const;
	uint16 MapOldOrder() const;
	void ConvertFromOldSavegame();
};

void InsertOrder(Vehicle *v, Order *new_o, VehicleOrderID sel_ord);
void DeleteOrder(Vehicle *v, VehicleOrderID sel_ord);

/**
 * Shared order list linking together the linked list of orders and the list
 *  of vehicles sharing this order list.
 */
struct OrderList : OrderListPool::PoolItem<&_orderlist_pool> {
// private:
	friend void AfterLoadVehicles(bool part_of_load); ///< For instantiating the shared vehicle chain
	friend SaveLoadTable GetOrderListDescription(); ///< Saving and loading of order lists.

	StationID GetBestLoadableNext(const Vehicle *v, const Order *o1, const Order *o2) const;

	Order *first;                     ///< First order of the order list.
	VehicleOrderID num_orders;        ///< NOSAVE: How many orders there are in the list.
	uint num_vehicles;                ///< NOSAVE: Number of vehicles that share this order list.
	Vehicle *first_shared;            ///< NOSAVE: pointer to the first vehicle in the shared order chain.

	TimeTick timetable_interval;      ///< Interval between beginning the order list.
	TimeTick timetable_duration;      ///< NOSAVE: Total timetabled duration of the order list.

//	std::vector<TimeTick> dispatch;
//	TimeTick dispatch_time;
//	TimeTick dispatch_max_delay;
//	TimeTick last_dispatch;

// public:
	/** Default constructor producing an invalid order list. */
	OrderList(VehicleOrderID num_orders = INVALID_VEH_ORDER_ID)
		: first(nullptr), num_orders(num_orders), num_vehicles(0), first_shared(nullptr),
		  timetable_interval(0), timetable_duration(0)
//		  dispatch_time(INVALID), dispatch_max_delay(0), last_dispatch(0)
	{ }

	/**
	 * Create an order list with the given order chain for the given vehicle.
	 *  @param chain pointer to the first order of the order chain
	 *  @param v any vehicle using this orderlist
	 */
	OrderList(Order *chain, Vehicle *v) { this->Initialize(chain, v); }

	/** Destructor. Invalidates OrderList for re-usage by the pool. */
	~OrderList() {}

	void Initialize(Order *chain, Vehicle *v);

	void RecalculateTimetableDuration();

	/**
	 * Get the first order of the order chain.
	 * @return the first order of the chain.
	 */
	Order *GetFirstOrder() const { return this->first; }

	Order *GetOrderAt(int index) const;

	/**
	 * Get the last order of the order chain.
	 * @return the last order of the chain.
	 */
	Order *GetLastOrder() const { return this->GetOrderAt(this->num_orders - 1); }

	/**
	 * Get the order after the given one or the first one, if the given one is the
	 * last one.
	 * @param curr Order to find the next one for.
	 * @return Next order.
	 */
	const Order *GetNext(const Order *curr) const { return (curr->next == nullptr) ? this->GetFirstOrder() : curr->next; }

	/**
	 * Get number of orders in the order list.
	 * @return number of orders in the chain.
	 */
	VehicleOrderID GetNumOrders() const { return this->num_orders; }

	StationIDStack GetNextStoppingStation(const Vehicle *v, const Order *first = nullptr, uint hops = 0) const;
	const Order *GetNextDecisionNode(const Order *next, uint hops) const;

	void InsertOrderAt(Vehicle *v, Order *new_order, int index);
	void DeleteOrderAt(Vehicle *v, int index);
	void MoveOrder(int from, int to);

	/**
	 * Is this a shared order list?
	 * @return whether this order list is shared among multiple vehicles
	 */
	bool IsShared() const { return this->num_vehicles > 1; };

	/**
	 * Get the first vehicle of this vehicle chain.
	 * @return the first vehicle of the chain.
	 */
	Vehicle *GetFirstSharedVehicle() const { return this->first_shared; }

	/**
	 * Return the number of vehicles that share this orders list
	 * @return the count of vehicles that use this shared orders list
	 */
	uint GetNumVehicles() const { return this->num_vehicles; }

	int GetPositionInSharedOrderList(const Vehicle *v) const;

	/**
	 * Adds the given vehicle to this shared order list.
	 * @note This is supposed to be called after the vehicle has been inserted
	 *       into the shared vehicle chain.
	 * @param v vehicle to add to the list
	 */
	void AddVehicle(Vehicle *v) { ++this->num_vehicles; }

	void RemoveVehicle(Vehicle *v);

//	bool IsCompleteTimetable() const;

	/**
	 * Gets the total duration of the vehicles timetable or INVALID_TICKS is the timetable is not complete.
	 * @return total timetable duration or INVALID_TICKS for incomplete timetables
	 */
//	TimeTick GetTimetableTotalDuration() const { return this->IsCompleteTimetable() ? this->timetable_duration : INVALID_TICKS; }

	/**
	 * Gets the known duration of the vehicles timetable even if the timetable is not complete.
	 * @return known timetable duration
	 */
//	TimeTick GetTimetableDurationIncomplete() const { return this->timetable_duration; }

	/**
	 * Must be called if an order's timetable is changed to update internal book keeping.
	 * @param delta By how many ticks has the timetable duration changed
	 */
//	void UpdateTimetableDuration(TimeTick delta) { this->timetable_duration += delta; }

	/**
     * Fetch the vector of all dispatch slots.
     * @return first dispatch time.
     */
//    const std::vector<TimeTick> &GetDispatch() const { return this->dispatch; }

//    void SetDispatch(std::vector<TimeTick> dispatch_list);
//    void TimetableAddTimeSlot(TimeTick offset);
//    void TimetableRemoveTimeSlot(TimeTick offset);
//    void UpdateDispatch();
//    void ResetDispatch();

    /**
     * Set a dispatch time
     * @param time = new time.
     */
//    void SetDispatchTime(TimeTick time) { this->dispatch_time = time; }

    /**
     * Fetch a dispatch time.
     * @return dispatch time.
     */
//    TimeTick GetDispatchTime() const { return this->dispatch_time; }

    /**
     * Whether the scheduled dispatch setting is valid.
     * @return bool.
     */
//    bool DispatchValid() const { return this->dispatch_time >= 0 && this->timetable_interval > 0; }

    /**
     * Set the dispatch last dispatch offset.
     * @param offset = new last dispatch offset.
     */
//    void SetLastDispatch(TimeTick offset) { this->last_dispatch = offset; }

    /**
     * Fetch the last dispatch offset.
     * @return last dispatch offset.
     */
//    TimeTick GetLastDispatch() const { return this->last_dispatch; }

    /**
     * Set the maximum allowed delay for a dispatch.
     * @param delay = maximum allowed delay.
     */
//    void SetDispatchDelay(TimeTick delay) { this->dispatch_max_delay = delay; }

    /**
     * Fetch the maximum alowed delay for a dispatch.
     * @return maximum allowed delay.
     */
//    TimeTick GetDispatchDelay() const { return this->dispatch_max_delay; }

	void FreeChain(bool keep_orderlist = false);

	void DebugCheckSanity() const;
};

bool ShouldStopAtStation(Vehicle *vehicle);

#endif /* ORDER_BASE_H */
