/*
 * This file is part of OpenTTD.
 * OpenTTD is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, version 2.
 * OpenTTD is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details. You should have received a copy of the GNU General Public License along with OpenTTD. If not, see <http://www.gnu.org/licenses/>.
 */

/** @file follow_track.hpp Template function for track followers */

#include "stdafx.h"
#include "follow_track_type.h"
#include "follow_track_rail.hpp"
#include "../pbs.h"
#include "../vehicle/roadveh.h"
#include "../station_base.h"
#include "../vehicle/train.h"
#include "../tunnelbridge.h"
#include "../tunnelbridge_map.h"
#include "../depot_map.h"
#include "pathfinder_func.h"

#include "safeguards.h"

/* begin static */

static void FollowTileExit          (CFollowTrackRail *ftrail);
static bool QueryNewTileTrackStatus (CFollowTrackRail *ftrail);
static bool CanEnterNewTile         (CFollowTrackRail *ftrail);
static bool TryReverse              (CFollowTrackRail *ftrail);

/* end static */

void CFollowTrackRail::Init(const Train *v, RailTypes railtype_override)
{
	m_veh = v;
	Init(v != nullptr ? v->owner : INVALID_OWNER, railtype_override == INVALID_RAILTYPES ? Train::From(v)->compatible_railtypes : railtype_override);
}

void CFollowTrackRail::Init(Owner o, RailTypes railtype_override)
{
	m_veh_owner = o;
	/* don't worry, all is inlined so compiler should remove unnecessary initializations */
	m_old_tile = INVALID_TILE;
	m_old_td = INVALID_TRACKDIR;
	m_new_tile = INVALID_TILE;
	m_new_td_bits = TRACKDIR_BIT_NONE;
	m_exitdir = INVALID_DIAGDIR;
	m_is_station = m_is_bridge = m_is_tunnel = false;
	m_tiles_skipped = 0;
	m_err = EC_NONE;
	m_railtypes = railtype_override;
	mask_reserved_tracks = true;
}

	/**
	 * main follower routine. Fills all members and return true on success.
	 *  Otherwise returns false if track can't be followed.
	 */
	bool Follow(CFollowTrackRail *ftrail, TileIndex old_tile, Trackdir old_td)
	{
		ftrail->m_old_tile = old_tile;
		ftrail->m_old_td = old_td;
		ftrail->m_err = EC_NONE;
		ftrail->m_exitdir = TrackdirToExitdir(ftrail->m_old_td);
		FollowTileExit(ftrail);
		if (!QueryNewTileTrackStatus(ftrail)) return TryReverse(ftrail);
		ftrail->m_new_td_bits &= DiagdirReachesTrackdirs(ftrail->m_exitdir);
		if (ftrail->m_new_td_bits == TRACKDIR_BIT_NONE || !CanEnterNewTile(ftrail)) {

			/* CanEnterNewTile already set a reason.
			 * Do NOT overwrite it (important for example for EC_RAIL_ROAD_TYPE).
			 * Only set a reason if CanEnterNewTile was not called */
			if (ftrail->m_new_td_bits == TRACKDIR_BIT_NONE) ftrail->m_err = EC_NO_WAY;

			return false;
		}
		if (GetTileRailType(ftrail->m_old_tile) != INVALID_RAILTYPE
				&& GetTileRailType(ftrail->m_new_tile) != INVALID_RAILTYPE
			) {
			ftrail->m_new_td_bits &= (TrackdirBits)~(int)TrackdirCrossesTrackdirs(ftrail->m_old_td);
			if (ftrail->m_new_td_bits == TRACKDIR_BIT_NONE) {
				ftrail->m_err = EC_90DEG;
				return false;
			}
		}
		return true;
	}

	bool MaskReservedTracks(CFollowTrackRail *ftrail)
	{
//		if (!ftrail->mask_reserved_tracks) return true;

		if (ftrail->m_is_station) {
			/* Check skipped station tiles as well. */
			TileIndexDiff diff = TileOffsByDiagDir(ftrail->m_exitdir);
			for (TileIndex tile = ftrail->m_new_tile - diff * ftrail->m_tiles_skipped; tile != ftrail->m_new_tile; tile += diff) {
				if (HasStationReservation(tile)) {
					ftrail->m_new_td_bits = TRACKDIR_BIT_NONE;
					ftrail->m_err = EC_RESERVED;
					return false;
				}
			}
		}

		TrackBits reserved = GetReservedTrackbits(ftrail->m_new_tile);
		/* Mask already reserved trackdirs. */
		ftrail->m_new_td_bits &= ~TrackBitsToTrackdirBits(reserved);
		/* Mask out all trackdirs that conflict with the reservation. */
		for (Track t : SetTrackBitIterator(TrackdirBitsToTrackBits(ftrail->m_new_td_bits))) {
			if (TracksOverlap(reserved | TrackToTrackBits(t))) ftrail->m_new_td_bits &= ~TrackToTrackdirBits(t);
		}
		if (ftrail->m_new_td_bits == TRACKDIR_BIT_NONE) {
			ftrail->m_err = EC_RESERVED;
			return false;
		}
		return true;
	}

	/** Follow the m_exitdir from m_old_tile and fill m_new_tile and m_tiles_skipped */
static void FollowTileExit(CFollowTrackRail *ftrail)
	{
		ftrail->m_is_station = ftrail->m_is_bridge = ftrail->m_is_tunnel = false;
		ftrail->m_tiles_skipped = 0;

		/* extra handling for tunnels and bridges in our direction */
		if (IsTileType(ftrail->m_old_tile, MP_TUNNELBRIDGE)) {
			DiagDirection enterdir = GetTunnelBridgeDirection(ftrail->m_old_tile);
			if (enterdir == ftrail->m_exitdir) {
				/* we are entering the tunnel / bridge */
				if (IsTunnel(ftrail->m_old_tile)) {
					ftrail->m_is_tunnel = true;
					ftrail->m_new_tile = GetOtherTunnelEnd(ftrail->m_old_tile);
				} else { // IsBridge(ftrail->m_old_tile)
					ftrail->m_is_bridge = true;
					ftrail->m_new_tile = GetOtherBridgeEnd(ftrail->m_old_tile);
				}
				ftrail->m_tiles_skipped = GetTunnelBridgeLength(ftrail->m_new_tile, ftrail->m_old_tile);
				return;
			}
			assert(ReverseDiagDir(enterdir) == ftrail->m_exitdir);
		}

		/* normal or station tile, do one step */
		ftrail->m_new_tile = TileAddByDiagDir(ftrail->m_old_tile, ftrail->m_exitdir);

		/* special handling for stations */
		if (HasStationTileRail(ftrail->m_new_tile)) {
			ftrail->m_is_station = true;
		}
	}

	/** stores track status (available trackdirs) for the new tile into m_new_td_bits */
static bool QueryNewTileTrackStatus(CFollowTrackRail *ftrail)
	{
		if (IsPlainRailTile(ftrail->m_new_tile)) {
			ftrail->m_new_td_bits = (TrackdirBits)(GetTrackBits(ftrail->m_new_tile) * 0x101);
		} else {
			ftrail->m_new_td_bits = TrackStatusToTrackdirBits(GetTileTrackStatus(ftrail->m_new_tile, TRANSPORT_RAIL, 0));
		}
		return (ftrail->m_new_td_bits != TRACKDIR_BIT_NONE);
	}

	/** return true if we can enter m_new_tile from m_exitdir */
static bool CanEnterNewTile(CFollowTrackRail *ftrail)
	{
		if (IsDepotTypeTile(ftrail->m_new_tile, TRANSPORT_RAIL)) {
			DiagDirection exitdir = GetRailDepotDirection(ftrail->m_new_tile);
			if (ReverseDiagDir(exitdir) != ftrail->m_exitdir) {
				ftrail->m_err = EC_NO_WAY;
				return false;
			}
		}

		/* rail transport is possible only on tiles with the same owner as vehicle */
		if (GetTileOwner(ftrail->m_new_tile) != ftrail->m_veh_owner) {
			/* different owner */
			ftrail->m_err = EC_NO_WAY;
			return false;
		}

		/* rail transport is possible only on compatible rail types */
		RailType rail_type = GetTileRailType(ftrail->m_new_tile);
		if (!HasBit(ftrail->m_railtypes, rail_type)) {
			/* incompatible rail type */
			ftrail->m_err = EC_RAIL_ROAD_TYPE;
			return false;
		}

		/* tunnel holes and bridge ramps can be entered only from proper direction */
		if (IsTileType(ftrail->m_new_tile, MP_TUNNELBRIDGE)) {
			if (IsTunnel(ftrail->m_new_tile)) {
				if (!ftrail->m_is_tunnel) {
					DiagDirection tunnel_enterdir = GetTunnelBridgeDirection(ftrail->m_new_tile);
					if (tunnel_enterdir != ftrail->m_exitdir) {
						ftrail->m_err = EC_NO_WAY;
						return false;
					}
				}
			} else { // IsBridge(ftrail->m_new_tile)
				if (!ftrail->m_is_bridge) {
					DiagDirection ramp_enderdir = GetTunnelBridgeDirection(ftrail->m_new_tile);
					if (ramp_enderdir != ftrail->m_exitdir) {
						ftrail->m_err = EC_NO_WAY;
						return false;
					}
				}
			}
		}

		/* special handling for rail stations - get to the end of platform */
		if (ftrail->m_is_station) {
			/* entered railway station
			 * get platform length */
			uint length = BaseStation::GetByTile(ftrail->m_new_tile)->GetPlatformLength(ftrail->m_new_tile, TrackdirToExitdir(ftrail->m_old_td));
			/* how big step we must do to get to the last platform tile; */
			ftrail->m_tiles_skipped = length - 1;
			/* move to the platform end */
			TileIndexDiff diff = TileOffsByDiagDir(ftrail->m_exitdir);
			diff *= ftrail->m_tiles_skipped;
			ftrail->m_new_tile = TILE_ADD(ftrail->m_new_tile, diff);
			return true;
		}

		return true;
	}

	/** return true if we successfully reversed at end of road/track */
static bool TryReverse(CFollowTrackRail *ftrail)
	{
		ftrail->m_err = EC_NO_WAY;
		return false;
	}

	/** Helper for pathfinders - get min/max speed on the m_old_tile/m_old_td */
	int GetSpeedLimit(const CFollowTrackRail *ftrail, int *pmin_speed = nullptr)
	{
		int min_speed = 0;
		int max_speed = INT_MAX; // no limit

		/* Check for speed limit imposed by railtype */
		uint16 rail_speed = GetRailTypeInfo(GetRailType(ftrail->m_old_tile))->max_speed;
		if (rail_speed > 0) max_speed = std::min<int>(max_speed, rail_speed);

		/* if min speed was requested, return it */
		if (pmin_speed != nullptr) *pmin_speed = min_speed;
		return max_speed;
	}
