/*
 * This file is part of OpenTTD.
 * OpenTTD is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, version 2.
 * OpenTTD is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details. You should have received a copy of the GNU General Public License along with OpenTTD. If not, see <http://www.gnu.org/licenses/>.
 */

/** @file follow_track.hpp Template function for track followers */

#include "follow_track_type.h"
#include "follow_track_road.hpp"
#include "pathfinder_func.h"

#include "../pbs.h"
#include "../station_base.h"
#include "../vehicle/roadveh.h"
#include "../vehicle/train.h"
#include "../tunnelbridge.h"
#include "../tunnelbridge_map.h"
#include "../depot_map.h"

/* begin static */

static DiagDirection GetSingleTramBit        (TileIndex tile);
static bool          ForcedReverse           (CFollowTrackRoad *ftroad);
static bool          CanExitOldTile          (CFollowTrackRoad *ftroad);
// static bool          MaskReservedTracks      (CFollowTrackRoad *ftroad);
static void          FollowTileExit          (CFollowTrackRoad *ftroad);
static bool          QueryNewTileTrackStatus (CFollowTrackRoad *ftroad);
static bool          CanEnterNewTile         (CFollowTrackRoad *ftroad);
static bool          TryReverse              (CFollowTrackRoad *ftroad);
static bool          IsTram                  (CFollowTrackRoad *ftroad);

/* end static */

	void CFollowTrackRoad::Init(const RoadVehicle *v, RailTypes railtype_override)
	{
		m_veh = v;
		Init(v != nullptr ? v->owner : INVALID_OWNER, railtype_override);
	}

	void CFollowTrackRoad::Init(Owner o, RailTypes railtype_override)
	{
		m_veh_owner = o;
		m_old_tile = INVALID_TILE;
		m_old_td = INVALID_TRACKDIR;
		m_new_tile = INVALID_TILE;
		m_new_td_bits = TRACKDIR_BIT_NONE;
		m_exitdir = INVALID_DIAGDIR;
		m_is_station = m_is_bridge = m_is_tunnel = false;
		m_tiles_skipped = 0;
		m_err = EC_NONE;
		m_railtypes = railtype_override;
	}

static bool IsTram(CFollowTrackRoad *ftroad)
{
	return RoadTypeIsTram(RoadVehicle::From(ftroad->m_veh)->roadtype);
}

/** Tests if a tile is a road tile with a single tramtrack (tram can reverse) */
static DiagDirection GetSingleTramBit(TileIndex tile)
	{
		assert(IsTram(ftroad)); // this function shouldn't be called in other cases

		if (IsNormalRoadTile(tile)) {
			RoadBits rb = GetRoadBits(tile, RTT_TRAM);
			switch (rb) {
				case ROAD_NW: return DIAGDIR_NW;
				case ROAD_SW: return DIAGDIR_SW;
				case ROAD_SE: return DIAGDIR_SE;
				case ROAD_NE: return DIAGDIR_NE;
				default: break;
			}
		}
		return INVALID_DIAGDIR;
	}

	/**
	 * main follower routine. Fills all members and return true on success.
	 *  Otherwise returns false if track can't be followed.
	 */
	bool Follow(CFollowTrackRoad *ftroad, TileIndex old_tile, Trackdir old_td)
	{
		ftroad->m_old_tile = old_tile;
		ftroad->m_old_td = old_td;
		ftroad->m_err = EC_NONE;
		assert(
			((TrackStatusToTrackdirBits(
				GetTileTrackStatus(ftroad->m_old_tile, TRANSPORT_ROAD, (ftroad->m_veh != nullptr) ? (ftroad->IsTram(ftroad) ? RTT_TRAM : RTT_ROAD) : 0)
			) & TrackdirToTrackdirBits(ftroad->m_old_td)) != 0) ||
			(IsTram(ftroad) && GetSingleTramBit(ftroad->m_old_tile) != INVALID_DIAGDIR) // Disable the assertion for single tram bits
		);
		ftroad->m_exitdir = TrackdirToExitdir(ftroad->m_old_td);
		if (ForcedReverse(ftroad)) return true;
		if (!CanExitOldTile(ftroad)) return false;
		FollowTileExit(ftroad);
		if (!QueryNewTileTrackStatus(ftroad)) return TryReverse(ftroad);
		ftroad->m_new_td_bits &= DiagdirReachesTrackdirs(ftroad->m_exitdir);
		if (ftroad->m_new_td_bits == TRACKDIR_BIT_NONE || !CanEnterNewTile(ftroad)) {
			/* In case we can't enter the next tile, but are
			 * a normal road vehicle, then we can actually
			 * try to reverse as this is the end of the road.
			 * Trams can only turn on the appropriate bits in
			 * which case reaching this would mean a dead end
			 * near a building and in that case there would
			 * a "false" QueryNewTileTrackStatus result and
			 * as such reversing is already tried. The fact
			 * that function failed can have to do with a
			 * missing road bit, or inability to connect the
			 * different bits due to slopes. */
			if (!IsTram(ftroad) && TryReverse(ftroad)) return true;

			/* CanEnterNewTile already set a reason.
			 * Do NOT overwrite it (important for example for EC_RAIL_ROAD_TYPE).
			 * Only set a reason if CanEnterNewTile was not called */
			if (ftroad->m_new_td_bits == TRACKDIR_BIT_NONE) ftroad->m_err = EC_NO_WAY;

			return false;
		}
		return true;
	}

//static bool MaskReservedTracks(CFollowTrackRoad *ftroad)
//	{
//		if (!DoTrackMasking()) return true;

//		if (ftroad->m_is_station) {
			/* Check skipped station tiles as well. */
//			TileIndexDiff diff = TileOffsByDiagDir(ftroad->m_exitdir);
//			for (TileIndex tile = ftroad->m_new_tile - diff * ftroad->m_tiles_skipped; tile != ftroad->m_new_tile; tile += diff) {
//				if (HasStationReservation(tile)) {
//					ftroad->m_new_td_bits = TRACKDIR_BIT_NONE;
//					ftroad->m_err = EC_RESERVED;
//					return false;
//				}
//			}
//		}

//		TrackBits reserved = GetReservedTrackbits(ftroad->m_new_tile);
//		/* Mask already reserved trackdirs. */
//		ftroad->m_new_td_bits &= ~TrackBitsToTrackdirBits(reserved);
//		/* Mask out all trackdirs that conflict with the reservation. */
//		for (Track t : SetTrackBitIterator(TrackdirBitsToTrackBits(ftroad->m_new_td_bits))) {
//			if (TracksOverlap(reserved | TrackToTrackBits(t))) ftroad->m_new_td_bits &= ~TrackToTrackdirBits(t);
//		}
//		if (ftroad->m_new_td_bits == TRACKDIR_BIT_NONE) {
//			ftroad->m_err = EC_RESERVED;
//			return false;
//		}
//		return true;
//	}

	/** Follow the ftroad->m_exitdir from ftroad->m_old_tile and fill ftroad->m_new_tile and ftroad->m_tiles_skipped */
static void FollowTileExit(CFollowTrackRoad *ftroad)
	{
		ftroad->m_is_station = ftroad->m_is_bridge = ftroad->m_is_tunnel = false;
		ftroad->m_tiles_skipped = 0;

		/* extra handling for tunnels and bridges in our direction */
		if (IsTileType(ftroad->m_old_tile, MP_TUNNELBRIDGE)) {
			DiagDirection enterdir = GetTunnelBridgeDirection(ftroad->m_old_tile);
			if (enterdir == ftroad->m_exitdir) {
				/* we are entering the tunnel / bridge */
				if (IsTunnel(ftroad->m_old_tile)) {
					ftroad->m_is_tunnel = true;
					ftroad->m_new_tile = GetOtherTunnelEnd(ftroad->m_old_tile);
				} else { // IsBridge(ftroad->m_old_tile)
					ftroad->m_is_bridge = true;
					ftroad->m_new_tile = GetOtherBridgeEnd(ftroad->m_old_tile);
				}
				ftroad->m_tiles_skipped = GetTunnelBridgeLength(ftroad->m_new_tile, ftroad->m_old_tile);
				return;
			}
			assert(ReverseDiagDir(enterdir) == ftroad->m_exitdir);
		}

		/* normal or station tile, do one step */
		ftroad->m_new_tile = TileAddByDiagDir(ftroad->m_old_tile, ftroad->m_exitdir);

		/* special handling for stations */
		if (IsRoadStopTile(ftroad->m_new_tile)) {
			ftroad->m_is_station = true;
		}
	}

	/** stores track status (available trackdirs) for the new tile into ftroad->m_new_td_bits */
static bool QueryNewTileTrackStatus(CFollowTrackRoad *ftroad)
	{
		ftroad->m_new_td_bits = GetTrackdirBitsForRoad(ftroad->m_new_tile, IsTram(ftroad) ? RTT_TRAM : RTT_ROAD);
		return (ftroad->m_new_td_bits != TRACKDIR_BIT_NONE);
	}

	/** return true if we can leave ftroad->m_old_tile in ftroad->m_exitdir */
static bool CanExitOldTile(CFollowTrackRoad *ftroad)
	{
		/* road stop can be left at one direction only unless it's a drive-through stop */
		if (IsStandardRoadStopTile(ftroad->m_old_tile)) {
			DiagDirection exitdir = GetRoadStopDir(ftroad->m_old_tile);
			if (exitdir != ftroad->m_exitdir) {
				ftroad->m_err = EC_NO_WAY;
				return false;
			}
		}

		/* single tram bits can only be left in one direction */
		if (IsTram(ftroad)) {
			DiagDirection single_tram = GetSingleTramBit(ftroad->m_old_tile);
			if (single_tram != INVALID_DIAGDIR && single_tram != ftroad->m_exitdir) {
				ftroad->m_err = EC_NO_WAY;
				return false;
			}
		}

		/* road depots can be also left in one direction only */
		if (IsDepotTypeTile(ftroad->m_old_tile, TRANSPORT_ROAD)) {
			DiagDirection exitdir = GetRoadDepotDirection(ftroad->m_old_tile);
			if (exitdir != ftroad->m_exitdir) {
				ftroad->m_err = EC_NO_WAY;
				return false;
			}
		}
		return true;
	}

	/** return true if we can enter ftroad->m_new_tile from ftroad->m_exitdir */
static bool CanEnterNewTile(CFollowTrackRoad *ftroad)
	{
		if (IsStandardRoadStopTile(ftroad->m_new_tile)) {
			/* road stop can be entered from one direction only unless it's a drive-through stop */
			DiagDirection exitdir = GetRoadStopDir(ftroad->m_new_tile);
			if (ReverseDiagDir(exitdir) != ftroad->m_exitdir) {
				ftroad->m_err = EC_NO_WAY;
				return false;
			}
		}

		/* single tram bits can only be entered from one direction */
		if (IsTram(ftroad)) {
			DiagDirection single_tram = GetSingleTramBit(ftroad->m_new_tile);
			if (single_tram != INVALID_DIAGDIR && single_tram != ReverseDiagDir(ftroad->m_exitdir)) {
				ftroad->m_err = EC_NO_WAY;
				return false;
			}
		}

		/* road and rail depots can also be entered from one direction only */
		if (IsDepotTypeTile(ftroad->m_new_tile, TRANSPORT_ROAD)) {
			DiagDirection exitdir = GetRoadDepotDirection(ftroad->m_new_tile);
			if (ReverseDiagDir(exitdir) != ftroad->m_exitdir) {
				ftroad->m_err = EC_NO_WAY;
				return false;
			}
			/* don't try to enter other company's depots */
			if (GetTileOwner(ftroad->m_new_tile) != ftroad->m_veh_owner) {
				ftroad->m_err = EC_OWNER;
				return false;
			}
		}

		/* road transport is possible only on compatible road types */
			const RoadVehicle *v = RoadVehicle::From(ftroad->m_veh);
			RoadType roadtype = GetRoadType(ftroad->m_new_tile, GetRoadTramType(v->roadtype));
			if (!HasBit(v->compatible_roadtypes, roadtype)) {
				/* incompatible road type */
				ftroad->m_err = EC_RAIL_ROAD_TYPE;
				return false;
			}

		/* tunnel holes and bridge ramps can be entered only from proper direction */
		if (IsTileType(ftroad->m_new_tile, MP_TUNNELBRIDGE)) {
			if (IsTunnel(ftroad->m_new_tile)) {
				if (!ftroad->m_is_tunnel) {
					DiagDirection tunnel_enterdir = GetTunnelBridgeDirection(ftroad->m_new_tile);
					if (tunnel_enterdir != ftroad->m_exitdir) {
						ftroad->m_err = EC_NO_WAY;
						return false;
					}
				}
			} else { // IsBridge(ftroad->m_new_tile)
				if (!ftroad->m_is_bridge) {
					DiagDirection ramp_enderdir = GetTunnelBridgeDirection(ftroad->m_new_tile);
					if (ramp_enderdir != ftroad->m_exitdir) {
						ftroad->m_err = EC_NO_WAY;
						return false;
					}
				}
			}
		}

		return true;
	}

	/** return true if we must reverse (in depots and single tram bits) */
static bool ForcedReverse(CFollowTrackRoad *ftroad)
	{
		/* rail and road depots cause reversing */
		if (IsDepotTypeTile(ftroad->m_old_tile, TRANSPORT_ROAD)) {
			DiagDirection exitdir = GetRoadDepotDirection(ftroad->m_old_tile);
			if (exitdir != ftroad->m_exitdir) {
				/* reverse */
				ftroad->m_new_tile = ftroad->m_old_tile;
				ftroad->m_new_td_bits = TrackdirToTrackdirBits(ReverseTrackdir(ftroad->m_old_td));
				ftroad->m_exitdir = exitdir;
				ftroad->m_tiles_skipped = 0;
				ftroad->m_is_tunnel = ftroad->m_is_bridge = ftroad->m_is_station = false;
				return true;
			}
		}

		/* Single tram bits and standard road stops cause reversing. */
		if (((IsTram(ftroad) && GetSingleTramBit(ftroad->m_old_tile) == ReverseDiagDir(ftroad->m_exitdir)) ||
				(IsStandardRoadStopTile(ftroad->m_old_tile) && GetRoadStopDir(ftroad->m_old_tile) == ReverseDiagDir(ftroad->m_exitdir)))) {
			/* reverse */
			ftroad->m_new_tile = ftroad->m_old_tile;
			ftroad->m_new_td_bits = TrackdirToTrackdirBits(ReverseTrackdir(ftroad->m_old_td));
			ftroad->m_exitdir = ReverseDiagDir(ftroad->m_exitdir);
			ftroad->m_tiles_skipped = 0;
			ftroad->m_is_tunnel = ftroad->m_is_bridge = ftroad->m_is_station = false;
			return true;
		}

		return false;
	}

	/** return true if we successfully reversed at end of road/track */
static bool TryReverse(CFollowTrackRoad *ftroad)
	{
		if (!IsTram(ftroad)) {
			/* if we reached the end of road, we can reverse the RV and continue moving */
			ftroad->m_exitdir = ReverseDiagDir(ftroad->m_exitdir);
			/* new tile will be the same as old one */
			ftroad->m_new_tile = ftroad->m_old_tile;
			/* set new trackdir bits to all reachable trackdirs */
			QueryNewTileTrackStatus(ftroad);
			ftroad->m_new_td_bits &= DiagdirReachesTrackdirs(ftroad->m_exitdir);
			if (ftroad->m_new_td_bits != TRACKDIR_BIT_NONE) {
				/* we have some trackdirs reachable after reversal */
				return true;
			}
		}
		ftroad->m_err = EC_NO_WAY;
		return false;
	}

	/** Helper for pathfinders - get min/max speed on the ftroad->m_old_tile/ftroad->m_old_td */
int GetSpeedLimit(CFollowTrackRoad *ftroad, int *pmin_speed = nullptr)
	{
		int min_speed = 0;
		int max_speed = INT_MAX; // no limit

		/* Check for on-bridge speed limit */
		if (IsBridgeTile(ftroad->m_old_tile)) {
			int spd = GetBridgeSpec(GetBridgeType(ftroad->m_old_tile))->speed;
			spd *= 2;
			max_speed = std::min(max_speed, spd);
		}
			/* max_speed is already in roadvehicle units, no need to further modify (divide by 2) */
			uint16 road_speed = GetRoadTypeInfo(GetRoadType(ftroad->m_old_tile, GetRoadTramType(RoadVehicle::From(ftroad->m_veh)->roadtype)))->max_speed;
			if (road_speed > 0) max_speed = std::min<int>(max_speed, road_speed);

		/* if min speed was requested, return it */
		if (pmin_speed != nullptr) *pmin_speed = min_speed;
		return max_speed;
	}
