/*
 * This file is part of OpenTTD.
 * OpenTTD is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, version 2.
 * OpenTTD is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details. You should have received a copy of the GNU General Public License along with OpenTTD. If not, see <http://www.gnu.org/licenses/>.
 */

/** @file follow_track.hpp Template function for track followers */

#include "follow_track_type.h"
#include "follow_track_water.hpp"
#include "../station_base.h"
#include "../tunnelbridge.h"
#include "../tunnelbridge_map.h"
#include "../depot_map.h"
#include "pathfinder_func.h"

/* begin static */

static void FollowTileExit          (CFollowTrackWater *ftwater);
static bool QueryNewTileTrackStatus (CFollowTrackWater *ftwater);
static bool CanEnterNewTile         (CFollowTrackWater *ftwater);
static bool TryReverse              (CFollowTrackWater *ftwater);
//	bool ForcedReverse();

/* end static */

	void CFollowTrackWater::Init(const Ship *v, RailTypes railtype_override)
	{
		m_veh = v;
		Init(v != nullptr ? v->owner : INVALID_OWNER, railtype_override);
	}

	void CFollowTrackWater::Init(Owner o, RailTypes railtype_override)
	{
		m_veh_owner = o;
		m_old_tile = INVALID_TILE;
		m_old_td = INVALID_TRACKDIR;
		m_new_tile = INVALID_TILE;
		m_new_td_bits = TRACKDIR_BIT_NONE;
		m_exitdir = INVALID_DIAGDIR;
		m_is_station = m_is_bridge = m_is_tunnel = false;
		m_tiles_skipped = 0;
		m_err = EC_NONE;
		m_railtypes = railtype_override;
	}

	/**
	 * main follower routine. Fills all members and return true on success.
	 *  Otherwise returns false if track can't be followed.
	 */
bool Follow(CFollowTrackWater *ftwater, TileIndex old_tile, Trackdir old_td)
	{
		ftwater->m_old_tile = old_tile;
		ftwater->m_old_td = old_td;
		ftwater->m_err = EC_NONE;
		ftwater->m_exitdir = TrackdirToExitdir(ftwater->m_old_td);
//		if (ForcedReverse()) return true;
		FollowTileExit(ftwater);
		if (!QueryNewTileTrackStatus(ftwater)) return TryReverse(ftwater);
		ftwater->m_new_td_bits &= DiagdirReachesTrackdirs(ftwater->m_exitdir);
		if (ftwater->m_new_td_bits == TRACKDIR_BIT_NONE || !CanEnterNewTile(ftwater)) {

			/* CanEnterNewTile already set a reason.
			 * Do NOT overwrite it (important for example for EC_RAIL_ROAD_TYPE).
			 * Only set a reason if CanEnterNewTile was not called */
			if (ftwater->m_new_td_bits == TRACKDIR_BIT_NONE) ftwater->m_err = EC_NO_WAY;

			return false;
		}
		return true;
	}

	/** Follow the ftwater->m_exitdir from ftwater->m_old_tile and fill ftwater->m_new_tile and ftwater->m_tiles_skipped */
static void FollowTileExit(CFollowTrackWater *ftwater)
	{
		ftwater->m_is_station = ftwater->m_is_bridge = ftwater->m_is_tunnel = false;
		ftwater->m_tiles_skipped = 0;

		/* extra handling for tunnels and bridges in our direction */
		if (IsTileType(ftwater->m_old_tile, MP_TUNNELBRIDGE)) {
			DiagDirection enterdir = GetTunnelBridgeDirection(ftwater->m_old_tile);
			if (enterdir == ftwater->m_exitdir) {
				/* we are entering the tunnel / bridge */
				if (IsTunnel(ftwater->m_old_tile)) {
					ftwater->m_is_tunnel = true;
					ftwater->m_new_tile = GetOtherTunnelEnd(ftwater->m_old_tile);
				} else { // IsBridge(ftwater->m_old_tile)
					ftwater->m_is_bridge = true;
					ftwater->m_new_tile = GetOtherBridgeEnd(ftwater->m_old_tile);
				}
				ftwater->m_tiles_skipped = GetTunnelBridgeLength(ftwater->m_new_tile, ftwater->m_old_tile);
				return;
			}
			assert(ReverseDiagDir(enterdir) == ftwater->m_exitdir);
		}

		/* normal or station tile, do one step */
		ftwater->m_new_tile = TileAddByDiagDir(ftwater->m_old_tile, ftwater->m_exitdir);
	}

	/** stores track status (available trackdirs) for the new tile into ftwater->m_new_td_bits */
static bool QueryNewTileTrackStatus(CFollowTrackWater *ftwater)
	{
		ftwater->m_new_td_bits = TrackStatusToTrackdirBits(GetTileTrackStatus(ftwater->m_new_tile, TRANSPORT_WATER, 0));
		return (ftwater->m_new_td_bits != TRACKDIR_BIT_NONE);
	}

	/** return true if we can enter ftwater->m_new_tile from ftwater->m_exitdir */
static bool CanEnterNewTile(CFollowTrackWater *ftwater)
	{
		/* tunnel holes and bridge ramps can be entered only from proper direction */
		if (IsTileType(ftwater->m_new_tile, MP_TUNNELBRIDGE)) {
			if (IsTunnel(ftwater->m_new_tile)) {
				if (!ftwater->m_is_tunnel) {
					DiagDirection tunnel_enterdir = GetTunnelBridgeDirection(ftwater->m_new_tile);
					if (tunnel_enterdir != ftwater->m_exitdir) {
						ftwater->m_err = EC_NO_WAY;
						return false;
					}
				}
			} else { // IsBridge(ftwater->m_new_tile)
				if (!ftwater->m_is_bridge) {
					DiagDirection ramp_enderdir = GetTunnelBridgeDirection(ftwater->m_new_tile);
					if (ramp_enderdir != ftwater->m_exitdir) {
						ftwater->m_err = EC_NO_WAY;
						return false;
					}
				}
			}
		}

		return true;
	}

	/** return true if we must reverse (in depots and single tram bits) */
/*	bool CFollowTrackWater::ForcedReverse()
	{
		return false;
	}
*/
	/** return true if we successfully reversed at end of road/track */
static bool TryReverse(CFollowTrackWater *ftwater)
	{
		ftwater->m_err = EC_NO_WAY;
		return false;
	}
