/*
 * This file is part of OpenTTD.
 * OpenTTD is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, version 2.
 * OpenTTD is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details. You should have received a copy of the GNU General Public License along with OpenTTD. If not, see <http://www.gnu.org/licenses/>.
 */

/** @file timetable_cmd.cpp Commands related to time tabling. */




#include "stdafx.h"

#include "timetable.h"

#include "cmd_helper.h"
#include "command_func.h"
#include "company_func.h"
#include "date_func.h"
#include "vehicle_base.h"
#include "window_func.h"

#include "table/strings.h"

#include "debug.h"
#include "safeguards.h"



/* begin static */

static void ChangeTimetable (
	Vehicle *v, VehicleOrderID order_number, TimeTick val, ModifyTimetableFlags mtf
);

static bool VehicleTimetableSorter (
	Vehicle * const &a, Vehicle * const &b
);

/* end static */





/**
 * Change/update a particular timetable entry.
 * @param v            The vehicle to change the timetable of.
 * @param order_number The index of the timetable in the order list.
 * @param val          The new data of the timetable entry.
 * @param mtf          Which part of the timetable entry to change.
 * @param timetabled   If the new value is explicitly timetabled.
 */
static void ChangeTimetable (
	Vehicle *v, VehicleOrderID order_number, TimeTick val, ModifyTimetableFlags mtf
) {

	Order *order = v->GetOrder(order_number);

	switch (mtf) {
		case MTF_SCHEDULED_ARRIVAL_TIME:
			order->scheduled_arrival = val;
			break;

		case MTF_SCHEDULED_DEPARTURE_TIME:
			order->scheduled_departure = val;
			break;

		case MTF_ARRIVAL_TIME:
			order->timed_arrival = val ;
			break;

		case MTF_DEPARTURE_TIME:
			order->timed_departure = val;
			break;

		case MTF_INTERVAL:
			v->orders.list->timetable_interval = val;
			break;

		case MTF_DELAY:
//			v->orders.list->SetDispatchDelay(val);
			break;

		case MTF_WAIT_TIME:
			order->wait_time = val;
			break;

		case MTF_TRAVEL_TIME:
			order->travel_time = val;
			break;

		case MTF_TRAVEL_SPEED:
			order->max_speed = val;
			break;

		default:
			NOT_REACHED();
	}
}





/**
 * Change timetable data of an order.
 * @param tile Not used.
 * @param flags Operation to perform.
 * @param p1 Various bitstuffed elements
 * - p1 = (bit  0-19) - Vehicle with the orders to change.
 * - p1 = (bit 20-27) - Order index to modify.
 * - p1 = (bit 28-29) - Timetable data to change (@see ModifyTimetableFlags)
 * @param p2 The amount of time to wait.
 * - p2 = (bit  0-15) - The data to modify as specified by p1 bits 28-29.
 *                      0 to clear times, UINT16_MAX to clear speed limit.
 * @param text unused
 * @return the cost of this operation or an error
 */
CommandCost CmdChangeTimetable (
	CommandContainer cmdvar
) {

	ProcessCmdFlag flags        = cmdvar.dc_flags;
	Vehicle *v                  = Vehicle::GetIfValid(cmdvar.var01);
	VehicleOrderID order_number = cmdvar.var02;
	ModifyTimetableFlags mtf    = (ModifyTimetableFlags)cmdvar.var03;
	TimeTick time               = cmdvar.var04;
	int16 speed                 = (int16)cmdvar.var05;
	bool clear                  = cmdvar.var08;

	if (v == nullptr || !v->IsPrimaryVehicle()) return CMD_ERROR;

	CommandCost ret = CheckOwnership(v->owner);
	if (ret.Failed()) return ret;

	Order *order = v->GetOrder(order_number);
	if (order == nullptr) return CMD_ERROR;

	if (mtf >= MTF_END) return CMD_ERROR;

	TimeTick scheduled_arrival   = order->scheduled_departure;
	TimeTick scheduled_departure = order->scheduled_departure;
	TimeTick travel_time         = order->travel_time;
	int max_speed                = order->max_speed;

	switch (mtf) {
		case MTF_TRAVEL_TIME:
			travel_time = clear ? 0 : time;
			break;

		case MTF_TRAVEL_SPEED:
			max_speed = clear ? UINT16_MAX : speed;
			if (max_speed == 0) max_speed = UINT16_MAX; // Disable speed limit.
			break;

		case MTF_WAIT_TIME:
		case MTF_SCHEDULED_ARRIVAL_TIME:
		case MTF_SCHEDULED_DEPARTURE_TIME:
		case MTF_ARRIVAL_TIME:
		case MTF_DEPARTURE_TIME:
		case MTF_INTERVAL:
		case MTF_DELAY:
			break;

		default:
			NOT_REACHED();
	}

	if (time != order->wait_time) {
		switch (order->GetType()) {
			case OT_GOTO_STATION:
				if (order->GetNonStopType() & ONSF_NO_STOP_AT_DESTINATION_STATION) return_cmd_error(STR_ERROR_TIMETABLE_NOT_STOPPING_HERE);
				break;

			case OT_GOTO_DEPOT:
			case OT_GOTO_WAYPOINT:
			case OT_CONDITIONAL:
				break;

			default: return_cmd_error(STR_ERROR_TIMETABLE_ONLY_WAIT_AT_STATIONS);
		}
	}

	if (travel_time != order->travel_time && order->IsType(OT_CONDITIONAL)) return CMD_ERROR;
	if (max_speed != order->max_speed && (order->IsType(OT_CONDITIONAL) || v->type == VEH_AIRCRAFT)) return CMD_ERROR;

	if (flags & DC_EXEC) {
		switch (mtf) {
			case MTF_SCHEDULED_ARRIVAL_TIME:
				/* Set time if changing the value or confirming an estimated time as timetabled. */
				if (time != order->scheduled_departure || (scheduled_arrival > 0)) {
					ChangeTimetable(v, order_number, time, MTF_SCHEDULED_ARRIVAL_TIME);
				}
				break;

			case MTF_SCHEDULED_DEPARTURE_TIME:
				/* Set time if changing the value or confirming an estimated time as timetabled. */
				if (time != order->scheduled_departure || (scheduled_departure > 0)) {
					ChangeTimetable(v, order_number, time, MTF_SCHEDULED_DEPARTURE_TIME);
				}
				break;

			case MTF_INTERVAL:
				/* Set time if changing the value or confirming an estimated time as timetabled. */
				ChangeTimetable(v, order_number, time, MTF_INTERVAL);
				break;

			case MTF_DELAY:
				/* Set time if changing the value or confirming an estimated time as timetabled. */
				ChangeTimetable(v, order_number, time, MTF_DELAY);
				break;

			case MTF_WAIT_TIME:
				/* Set time if changing the value or confirming an estimated time as timetabled. */
				if (time != order->wait_time) {
					ChangeTimetable(v, order_number, time, MTF_WAIT_TIME);
				}
				break;

//			case MTF_TRAVEL_TIME:
				/* Set time if changing the value or confirming an estimated time as timetabled. */
//				if (travel_time != order->travel_time || (travel_time > 0 && !order->IsTravelTimetabled())) {
//					ChangeTimetable(v, order_number, travel_time, MTF_TRAVEL_TIME);
//				}
//				break;

			case MTF_TRAVEL_SPEED:
				if (max_speed != order->max_speed) {
					ChangeTimetable(v, order_number, max_speed, MTF_TRAVEL_SPEED);
				}
				break;

			default:
				break;
		}

		if (v->vehicle_flags & VF_SHARED_ORDERS) {
			for (Vehicle *u = v->FirstShared(); u != nullptr; u = u->NextShared()) {
				switch (mtf) {
					case MTF_WAIT_TIME:
							ChangeTimetable(u, order_number, time, MTF_WAIT_TIME);
						break;

					case MTF_TRAVEL_SPEED:
							ChangeTimetable(u, order_number, max_speed, MTF_TRAVEL_SPEED);
						break;

					default:
						break;
				}
			}
		}
	}

	return CommandCost();
}





/**
 * Clear the lateness counter to make the vehicle on time.
 * @param tile Not used.
 * @param flags Operation to perform.
 * @param p1 Various bitstuffed elements
 * - p1 = (bit  0-19) - Vehicle with the orders to change.
 * @param p2 unused
 * @param text unused
 * @return the cost of this operation or an error
 */
CommandCost CmdSetVehicleOnTime (
	CommandContainer cmdvar
) {

	ProcessCmdFlag flags = cmdvar.dc_flags;
	Vehicle *v          = Vehicle::GetIfValid(cmdvar.var01);

	if (v == nullptr || !v->IsPrimaryVehicle() || v->orders.list == nullptr) return CMD_ERROR;

	CommandCost ret = CheckOwnership(v->owner);
	if (ret.Failed()) return ret;

	if (flags & DC_EXEC) {
		v->lateness_counter = 0;
		SetWindowDirty(WC_VEHICLE_TIMETABLE, v->index);
	}

	return CommandCost();
}





/**
 * Order vehicles based on their timetable. The vehicles will be sorted in order
 * they would reach the first station.
 *
 * @param a First Vehicle pointer.
 * @param b Second Vehicle pointer.
 * @return Comparison value.
 */
static bool VehicleTimetableSorter (
	Vehicle * const &a, Vehicle * const &b
) {

	VehicleOrderID a_order = a->cur_real_order_index;
	VehicleOrderID b_order = b->cur_real_order_index;
	int j = (int)b_order - (int)a_order;

	/* Are we currently at an ordered station (un)loading? */
	bool a_load = a->current_order.IsType(OT_LOADING) && a->current_order.GetNonStopType() != ONSF_STOP_EVERYWHERE;
	bool b_load = b->current_order.IsType(OT_LOADING) && b->current_order.GetNonStopType() != ONSF_STOP_EVERYWHERE;

	/* If the current order is not loading at the ordered station, decrease the order index by one since we have
	 * not yet arrived at the station (and thus the timetable entry; still in the travelling of the previous one).
	 * Since the ?_order variables are unsigned the -1 will flow under and place the vehicles going to order #0 at
	 * the begin of the list with vehicles arriving at #0. */
	if (!a_load) a_order--;
	if (!b_load) b_order--;

	/* First check the order index that accounted for loading, then just the raw one. */
	int i = (int)b_order - (int)a_order;
	if (i != 0) return i < 0;
	if (j != 0) return j < 0;

	/* Look at the time we spent in this order; the higher, the closer to its destination. */
	TimeTick time_diff = b->current_order_time - a->current_order_time;
	if (time_diff != 0) return i < 0;

	/* If all else is equal, use some unique index to sort it the same way. */
	return b->unitnumber < a->unitnumber;
}





/**
 * Set the start date of the timetable.
 * @param tile Not used.
 * @param flags Operation to perform.
 * @param p2 Various bitstuffed elements
 * - p2 = (bit 0-19) - Vehicle ID.
 * - p2 = (bit 20)   - Set to 1 to set timetable start for all vehicles sharing this order
 * @param p2 The timetable start date.
 * @param text Not used.
 * @return The error or cost of the operation.
 */
CommandCost CmdSetTimetableStart (
	CommandContainer cmdvar
) {

	ProcessCmdFlag flags = cmdvar.dc_flags;
	Vehicle *v          = Vehicle::GetIfValid(cmdvar.var01);
	TimeTick start_time = cmdvar.var04;
	bool timetable_all  = cmdvar.var08;

	if (v == nullptr || !v->IsPrimaryVehicle() || v->orders.list == nullptr) return CMD_ERROR;

	CommandCost ret = CheckOwnership(v->owner);
	if (ret.Failed()) return ret;

	if (!_enable_clock) {
		/* Don't let a timetable start more than 15 years into the future or 1 year in the past. */
		if (start_time < 0 || start_time > ((TimeTick)MAX_DAY * DAY_LENGTH))             return CMD_ERROR;
		if (start_time - (_date * DAY_LENGTH) > (15ll * DAYS_IN_LEAP_YEAR * DAY_LENGTH)) return CMD_ERROR;
		if (_date * DAY_LENGTH - start_time > DAYS_IN_LEAP_YEAR * DAY_LENGTH)            return CMD_ERROR;
//		if (timetable_all && !v->orders.list->IsCompleteTimetable())                     return CMD_ERROR;
	}

	if (flags & DC_EXEC) {
		std::vector<Vehicle *> vehs;

		if (timetable_all) {
			for (Vehicle *w = v->orders.list->GetFirstSharedVehicle(); w != nullptr; w = w->NextShared()) {
				vehs.push_back(w);
			}
		} else {
			vehs.push_back(v);
		}

//		TimeTick total_duration = v->orders.list->GetTimetableTotalDuration();
		TimeTick total_duration = 1;
		int num_vehs = (uint)vehs.size();

		if (num_vehs >= 2) {
			std::sort(vehs.begin(), vehs.end(), &VehicleTimetableSorter);
		}

		int idx = vehs.begin() - std::find(vehs.begin(), vehs.end(), v);

		for (Vehicle *w : vehs) {

			w->lateness_counter = 0;
			w->vehicle_flags &= ~VF_TIMETABLE_STARTED;
			/* Do multiplication, then division to reduce rounding errors. */
//o			w->timetable_start = start_time + idx * total_duration / num_vehs / DAY_LENGTH;
			w->timetable_start = start_time + idx * total_duration / num_vehs;
//			v->orders.list->SetDispatchTime( start_time + idx * total_duration / num_vehs);

			SetWindowDirty(WC_VEHICLE_TIMETABLE, w->index);
			++idx;
		}

	}

	return CommandCost();
}





//
// Simpler form? Use dispatch flag and array of departure times updated with interval?
//

/**
 * Add scheduled dispatch time offset
 * @param tile Not used.
 * @param flags Operation to perform.
 * @param p1 Vehicle index.
 * @param p2 Offset time to add.
 * @param text unused
 * @return the cost of this operation or an error
 */
/*
CommandCost CmdTimetableAddTimeSlot(CommandContainer cmdvar)
{
	ProcessCmdFlag dc_flags = cmdvar.dc_flags;
	Vehicle *v             = Vehicle::GetIfValid(cmdvar.var01);
	TimeTick time          = cmdvar.var02;

	if (v == nullptr || !v->IsPrimaryVehicle() || v->orders.list == nullptr) return CMD_ERROR;

	CommandCost ret = CheckOwnership(v->owner);
	if (ret.Failed()) return ret;

	if (dc_flags & DC_EXEC) {
		v->orders.list->TimetableAddTimeSlot(time);
//		SetWindowDirty(WC_SCHDISPATCH_SLOTS, v->index);
	}

	return CommandCost();
}
*/

/**
 * Set scheduled dispatch slot list.
 * @param dispatch_list The offset time list, must be correctly sorted.
 */
/*
void OrderList::SetDispatch(std::vector<TimeTick> dispatch_list)
{
	this->dispatch = std::move(dispatch_list);
	assert(std::is_sorted(this->dispatch.begin(), this->dispatch.end()));
	if (this->DispatchValid()) this->UpdateDispatch();
}
*/
/**
 * Add new scheduled dispatch slot at offsets time.
 * @param offset The offset time to add.
 */
/*
void OrderList::TimetableAddTimeSlot(TimeTick offset)
{
	/* Maintain sorted list status */
/*	auto insert_position = std::lower_bound(this->dispatch.begin(), this->dispatch.end(), offset);
	if (insert_position != this->dispatch.end() && *insert_position == offset) {
		return;
	}
	this->dispatch.insert(insert_position, offset);
	this->UpdateDispatch();
}
*/
/**
 * Remove scheduled dispatch slot at offsets time.
 * @param offset The offset time to remove.
 */
/*void OrderList::TimetableRemoveTimeSlot(TimeTick offset)
{
	/* Maintain sorted list status */
/*	auto erase_position = std::lower_bound(this->dispatch.begin(), this->dispatch.end(), offset);
	if (erase_position == this->dispatch.end() || *erase_position != offset) {
		return;
	}
	this->dispatch.erase(erase_position);
}

/**
 * Update the scheduled dispatch start time to be the most recent possible.
 */
/*void OrderList::UpdateDispatch()
{
	if (this->timetable_interval == 0 || this->GetDispatchTime() < 0) return;
	if (this->last_dispatch == 0) this->last_dispatch = _time_tick;
//	bool update_windows = false;
	/* Most of the time this loop does not runs. It makes sure start date in in past */
//	while (this->GetDispatchTime() > _time_tick) {
//		this->last_dispatch += this->GetTimetableInterval();
//		update_windows = true;
//	}
	/* Most of the time this loop runs once. It makes sure the start date is as close to current time as possible. */
//	while (this->GetDispatchTime() + this->GetTimetableInterval() <= _time_tick) {
//		this->last_dispatch -= this->GetTimetableInterval();
//		update_windows = true;
//	}
//	if (update_windows) InvalidateWindowClassesData(WC_TIMETABLE_TIME_SLOTS, VIWD_MODIFY_ORDERS);
//}





/**
 * Start or stop filling the timetable automatically from the time the vehicle
 * actually takes to complete it. When starting to autofill the current times
 * are cleared and the timetable will start again from scratch.
 * @param tile Not used.
 * @param flags Operation to perform.
 * @param p1 Vehicle index.
 * @param p2 Various bitstuffed elements
 * - p2 = (bit 0) - Set to 1 to enable, 0 to disable autofill.
 * - p2 = (bit 1) - Set to 1 to preserve waiting times in non-destructive mode
 * @param text unused
 * @return the cost of this operation or an error
 */
CommandCost CmdAutofillTimetable (
	CommandContainer cmdvar
) {

	ProcessCmdFlag flags = cmdvar.dc_flags;
	Vehicle *v           = Vehicle::GetIfValid(cmdvar.var01);
	bool autofill        = v->vehicle_flags & VF_AUTOFILL_TIMETABLE;
	bool preserve        = cmdvar.var09;

	if (v == nullptr || !v->IsPrimaryVehicle() || v->orders.list == nullptr) return CMD_ERROR;

	CommandCost ret = CheckOwnership(v->owner);
	if (ret.Failed()) return ret;

	if (flags & DC_EXEC) {
		if (!autofill) {
			/* Start autofilling the timetable, which clears the
			 * "timetable has started" bit. Times are not cleared anymore, but are
			 * overwritten when the order is reached now. */
			v->vehicle_flags |= VF_AUTOFILL_TIMETABLE;
			v->vehicle_flags ^= VF_TIMETABLE_STARTED;

			/* Overwrite waiting times only if they became longer */
			if (preserve) v->vehicle_flags |= VF_AUTOFILL_PRES_WAIT_TIME;

			v->timetable_start = 0;
			v->lateness_counter = 0;
		} else {
			v->vehicle_flags &= ~VF_AUTOFILL_TIMETABLE;
			v->vehicle_flags &= ~VF_AUTOFILL_PRES_WAIT_TIME;
		}

		for (Vehicle *v2 = v->FirstShared(); v2 != nullptr; v2 = v2->NextShared()) {
			if (v2 != v) {
				/* Stop autofilling; only one vehicle at a time can perform autofill */
				v2->vehicle_flags &= ~VF_AUTOFILL_TIMETABLE;
				v2->vehicle_flags &= ~VF_AUTOFILL_PRES_WAIT_TIME;
			}
			SetWindowDirty(WC_VEHICLE_TIMETABLE, v2->index);
		}
	}

	return CommandCost();
}





/**
 * Update the timetable for the vehicle.
 * @param v The vehicle to update the timetable for.
 * @param travelling Whether we just travelled or waited at a station.
 */
void UpdateVehicleTimetable (
	Vehicle *v, TimetableFlag flag
) {

	if (v->cur_real_order_index >= v->GetNumOrders()) return;
	Order *real_current_order = v->GetOrder(v->cur_real_order_index);

	/* Set arrival time. */
	if (!real_current_order->IsType(OT_CONDITIONAL) && flag == ARRIVAL) {
		ChangeTimetable(v, v->cur_real_order_index, _time_tick, MTF_ARRIVAL_TIME);
		TimeTick scheduled_time = real_current_order->scheduled_arrival;

		if (scheduled_time != 0) {
			TimeTick interval = v->orders.list->timetable_interval;

			if (interval != 0) {
				do {
					scheduled_time += interval;
				} while (scheduled_time < _time_tick);
				real_current_order->scheduled_arrival = scheduled_time;
			}

			else {
				real_current_order->scheduled_arrival = 0;
			}
		}
	}

	/* Set departure time. */
	if (!real_current_order->IsType(OT_CONDITIONAL) && flag == DEPARTURE) {
		ChangeTimetable(v, v->cur_real_order_index, _time_tick, MTF_DEPARTURE_TIME);
		TimeTick scheduled_time = real_current_order->scheduled_departure;
		if (scheduled_time != 0) {
			TimeTick interval = v->orders.list->timetable_interval;
			if (interval != 0) {
				do {
					scheduled_time += interval;
				} while (scheduled_time < _time_tick);
				real_current_order->scheduled_departure = scheduled_time;
			}
			else {
				real_current_order->scheduled_departure = 0;
			}
		}
	}
}
