/*
 * This file is part of OpenTTD.
 * OpenTTD is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, version 2.
 * OpenTTD is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details. You should have received a copy of the GNU General Public License along with OpenTTD. If not, see <http://www.gnu.org/licenses/>.
 */

/** @file timetable_gui.cpp GUI for time tabling. */

#include "stdafx.h"
#include "command_func.h"
#include "gui.h"
#include "window_gui.h"
#include "window_func.h"
#include "textbuf_gui.h"
#include "strings_func.h"
#include "vehicle_base.h"
#include "string_func.h"
#include "gfx_func.h"
#include "company_func.h"
#include "date_func.h"
#include "date_gui.h"
#include "vehicle_gui.h"
#include "settings_type.h"
#include "timetable.h"

#include "widgets/timetable_widget.h"

#include "spritecache.h"
#include "table/sprites.h"
#include "table/strings.h"

#include "debug.h"
#include "safeguards.h"

/**
 * Set the timetable parameters in the format as described by the setting.
 * @param param1 the first DParam to fill
 * @param param2 the second DParam to fill
 * @param ticks  the number of ticks to 'draw'
 */
void SetTimetableParams(int param1, int param2, TimeTick ticks)
{
	if (_settings_client.gui.timetable_in_ticks) {
		SetDParam(param1, STR_TIMETABLE_TICKS);
		SetDParam(param2, ticks);
	} else if (_enable_clock) {
		SetDParam(param1, STR_TIMETABLE_MINUTES);
		SetDParam(param2, ticks / MINUTE);
	} else {
		SetDParam(param1, STR_TIMETABLE_DAYS);
		SetDParam(param2, ticks / DAY_LENGTH);
	}
}

/**
 * Fill the table with arrivals and departures
 * @param v Vehicle which must have at least 2 orders.
 * @param start order index to start at
 * @param travelling Are we still in the travelling part of the start order
 * @param table Fill in arrival and departures including intermediate orders
 * @param offset Add this value to result and all arrivals and departures
 */
static void FillTimetableArrivalDepartureTable(const Vehicle *v, TimetableArrivalDeparture *table)
{
	bool travelling = (!v->current_order.IsType(OT_LOADING) || v->current_order.GetNonStopType() == ONSF_STOP_EVERYWHERE);
	VehicleOrderID current_order = v->cur_real_order_index % v->GetNumOrders();

	assert(table != nullptr);
	assert(v->GetNumOrders() >= 2);
	assert(current_order < v->GetNumOrders());

	VehicleOrderID order_id = current_order;
	const Order *order = v->GetOrder(order_id);

	/* Pre-initialize with unknown time */
	for (int i = 0; i < v->GetNumOrders(); ++i) {
		table[i].scheduled_arrival = table[i].scheduled_departure = table[i].arrival = table[i].departure = INVALID_TICKS;
	}

	/* Cyclically loop over all orders until we reach the current one again.
	 * As we may start at the current order, do a post-checking loop */
	do {
		/* List times also for conditional orders because otherwise it messes up the table. */
		table[order_id].scheduled_arrival = order->scheduled_arrival;
		table[order_id].scheduled_departure = order->scheduled_departure;
		table[order_id].arrival = order->timed_arrival;
		table[order_id].departure = order->timed_departure;

		++order_id;
		order = order->next;
		if (order_id >= v->GetNumOrders()) {
			order_id = 0;
			assert(order == nullptr);
			order = v->orders.list->GetFirstOrder();
		}
	} while (order_id != current_order);
}


/**
 * Callback for when a time has been chosen to start the time table
 * @param w the window related to the setting of the date
 * @param date the actually chosen date
 */
void ChangeTimetableStartCallback(const Window *w, TimeTick tick)
{
	CommandContainer cmdcont {};
		cmdcont.cmd   = CMD_SET_TIMETABLE_START | CMD_MSG(STR_ERROR_CAN_T_TIMETABLE_VEHICLE);
		cmdcont.var01 = w->window_number;
		cmdcont.var02 = tick;
	ProcessCmdP(cmdcont);
}

/**
 * Callback for when a time has been chosen to add to the schedule
 * @param p1 The p1 parameter to send to CmdScheduledDispatchAdd
 * @param date the actually chosen date
 */
/*
void TimetableAddTimeSlot(uint32 veh, TimeTick time)
{
	Vehicle *v = Vehicle::GetIfValid(veh);
	if (v == nullptr || !v->IsPrimaryVehicle()) return;

	CommandContainer cmdcont {};
		cmdcont.cmd   = CMD_TIMETABLE_ADD_TIME_SLOT | CMD_MSG(STR_ERROR_CAN_T_TIMETABLE_VEHICLE);
		cmdcont.var01 = v->index;
		cmdcont.var02 = time;
	ProcessCmdP(cmdcont);
}
*/
struct TimetableWindow : Window {
	int sel_index;
	const Vehicle *vehicle;    ///< Vehicle monitored by the window.
	int clicked_widget;        ///< The widget that was clicked (used to determine what to do in OnQueryTextFinished)
	bool show_scheduled;       ///< Whether we show expected arrival or scheduled
	bool set_start_date_all;   ///< Set start date for all vehicles sharing these orders.
	uint deparr_time_width;    ///< The width of the departure/arrival time
	uint deparr_abbr_width;    ///< The width of the departure/arrival abbreviation
	uint num_columns;          ///< Number of columns.
	uint item_count = 0;       ///< Number of scheduled item

	Scrollbar *vscroll;
	uint flag_width;
	uint flag_height;
	uint base_width;
	uint header_width;

	TimetableWindow(WindowDesc *desc, WindowNumber window_number) :
			Window(desc),
			sel_index(-1),
			vehicle(Vehicle::Get(window_number)),
			show_scheduled(true)
	{
		this->CreateNestedTree();
		this->vscroll = this->GetScrollbar(WID_VT_SCROLLBAR);
		this->UpdateSelectionStates();
		this->FinishInitNested(window_number);
		this->owner = this->vehicle->owner;
	}

	/**
	 * Build the arrival-departure list for a given vehicle
	 * @param v the vehicle to make the list for
	 * @param table the table to fill
	 * @return if next arrival will be early
	 */
	static bool BuildArrivalDepartureList(const Vehicle *v, TimetableArrivalDeparture *table)
	{
		if (v->GetNumOrders() > 0) FillTimetableArrivalDepartureTable(v, table);
		bool travelling = (!v->current_order.IsType(OT_LOADING)
							|| v->current_order.GetNonStopType() == ONSF_STOP_EVERYWHERE);
		return (travelling && v->lateness_counter < 0LL);
	}

	void UpdateWidgetSize(int widget, Dimension *size, const Dimension &padding, Dimension *fill, Dimension *resize) override
	{
		switch (widget) {
			case WID_VT_ARRIVAL_DEPARTURE_PANEL: {
				if (_enable_clock) {
					SetDParamMaxValue(0, UINT64_MAX, 0, FS_SMALL);
					this->deparr_time_width = GetStringBoundingBox(STR_TIME).width + 40;
				} else {
					SetDParamMaxValue(0, MAX_YEAR * DAYS_IN_YEAR, 0, FS_SMALL);
					this->deparr_time_width = GetStringBoundingBox(STR_JUST_DATE_TINY).width;
				}
				this->deparr_abbr_width = std::max(GetStringBoundingBox(STR_TIMETABLE_ARRIVAL_ABBREVIATION).width, GetStringBoundingBox(STR_TIMETABLE_DEPARTURE_ABBREVIATION).width);
				size->width = WD_FRAMERECT_LEFT + this->deparr_abbr_width + 10 + this->deparr_time_width + WD_FRAMERECT_RIGHT;
				FALLTHROUGH;
			}

			case WID_VT_TIMES_PANEL_SELECTION:
			case WID_VT_TIMETABLE_PANEL: {
				resize->height = FONT_HEIGHT_NORMAL;
				size->height = WD_FRAMERECT_TOP + 8 * resize->height + WD_FRAMERECT_BOTTOM;
				break;
			}

/*			case WID_VT_TIME_SLOT_MATRIX: {
				Dimension unumber;
				uint min_height = 0;
				if (_enable_clock) {
					SetDParamMaxValue(0, INT64_MAX, 0, FS_SMALL);
					unumber = GetStringBoundingBox(STR_TIME).width + 40;
				} else {
					SetDParamMaxValue(0, MAX_YEAR * DAYS_IN_YEAR, 0, FS_SMALL);
					unumber = GetStringBoundingBox(STR_JUST_DATE_TINY).width;
				}
				const Sprite *spr = GetSprite(SPR_FLAG_VEH_STOPPED, ST_NORMAL);
				this->flag_width  = spr->width + WD_FRAMERECT_RIGHT;
				this->flag_height = spr->height;

				min_height = std::max<uint>(unumber.height + WD_MATRIX_TOP, spr->height);
				this->header_width = this->flag_width + WD_FRAMERECT_LEFT;
				this->base_width = unumber.width + this->header_width + 4;

				resize->height = min_height;
				resize->width = base_width;
				size->width = resize->width * 3;
				size->height = resize->height * 1;

				fill->width = resize->width;
				fill->height = resize->height;
				break;
			}
*/

			case WID_VT_SUMMARY_PANEL: {
				size->height = WD_FRAMERECT_TOP + 2 * FONT_HEIGHT_NORMAL + WD_FRAMERECT_BOTTOM;
				break;
			}
		}
	}

	int GetOrderFromTimetableWndPt(int y, const Vehicle *v)
	{
		uint sel = (y - this->GetWidget<NWidgetBase>(WID_VT_TIMETABLE_PANEL)->pos_y - WD_FRAMERECT_TOP) / FONT_HEIGHT_NORMAL;

		if (sel >= this->vscroll->GetCapacity()) return INVALID_ORDER;

		sel += this->vscroll->GetPosition();

		return (sel < v->GetNumOrders() * 2u) ? sel : INVALID_ORDER;
	}

	/**
	 * Some data on this window has become invalid.
	 * @param data Information about the changed data.
	 * @param gui_scope Whether the call is done from GUI scope. You may not do everything when not in GUI scope. See #InvalidateWindowData() for details.
	 */
	void OnInvalidateData(int data = 0, bool gui_scope = true) override
	{
		switch (data) {
			case VIWD_AUTOREPLACE:
				/* Autoreplace replaced the vehicle */
				this->vehicle = Vehicle::Get(this->window_number);
				break;

			case VIWD_REMOVE_ALL_ORDERS:
				/* Removed / replaced all orders (after deleting / sharing) */
				if (this->sel_index == -1) break;

				this->CloseChildWindows();
				this->sel_index = -1;
				break;

			case VIWD_MODIFY_ORDERS:
				if (!gui_scope) break;
				this->UpdateSelectionStates();
				this->ReInit();
				break;

			default: {
				if (gui_scope) break; // only do this once; from command scope

				/* Moving an order. If one of these is INVALID_VEH_ORDER_ID, then
				 * the order is being created / removed */
				if (this->sel_index == -1) break;

				VehicleOrderID from = GB(data, 0, 8);
				VehicleOrderID to   = GB(data, 8, 8);

				if (from == to) break; // no need to change anything

				/* if from == INVALID_VEH_ORDER_ID, one order was added; if to == INVALID_VEH_ORDER_ID, one order was removed */
				uint old_num_orders = this->vehicle->GetNumOrders() - (uint)(from == INVALID_VEH_ORDER_ID) + (uint)(to == INVALID_VEH_ORDER_ID);

				VehicleOrderID selected_order = this->sel_index / 2;
				if (selected_order == old_num_orders) selected_order = 0; // when last travel time is selected, it belongs to order 0

				bool travel = HasBit(this->sel_index, 0);

				if (from != selected_order) {
					/* Moving from preceding order? */
					selected_order -= (int)(from <= selected_order);
					/* Moving to   preceding order? */
					selected_order += (int)(to   <= selected_order);
				} else {
					/* Now we are modifying the selected order */
					if (to == INVALID_VEH_ORDER_ID) {
						/* Deleting selected order */
						this->CloseChildWindows();
						this->sel_index = -1;
						break;
					} else {
						/* Moving selected order */
						selected_order = to;
					}
				}

				/* recompute new sel_index */
				this->sel_index = 2 * selected_order - (int)travel;
				/* travel time of first order needs special handling */
				if (this->sel_index == -1) this->sel_index = this->vehicle->GetNumOrders() * 2 - 1;
				break;
			}
		}
	}


	void OnPaint() override
	{
		const Vehicle *v = this->vehicle;
		int selected = this->sel_index;
//		CountItem();

		this->vscroll->SetCount(v->GetNumOrders() * 2);

		if (v->owner == _local_company) {
			bool disable = true;
			if (selected != -1) {
				const Order *order = v->GetOrder(((selected + 1) / 2) % v->GetNumOrders());
				if (selected % 2 == 1) {
//o					disable = order != nullptr && (order->IsType(OT_CONDITIONAL) || order->IsType(OT_IMPLICIT));
//o				} else {
//o					disable = order == nullptr || (((!order->IsType(OT_GOTO_STATION) || !order->IsType(OT_GOTO_DEPOT)) || (order->GetNonStopType() & ONSF_NO_STOP_AT_DESTINATION_STATION)) && !order->IsType(OT_CONDITIONAL));
                    /* Travel time */
                    disable = order != nullptr && (order->IsType(OT_CONDITIONAL));
                } else {
                    /* Wait time */
                    if (order != nullptr) {
                        if (order->IsType(OT_GOTO_WAYPOINT)) {
                            disable = false;
                        } else if (order->IsType(OT_CONDITIONAL)) {
                            disable = true;
                        } else {
//o							disable = (!(order->IsType(OT_GOTO_STATION) || (order->IsType(OT_GOTO_DEPOT) && !(order->GetDepotActionType() & ODATFB_HALT))) || (order->GetNonStopType() & ONSF_NO_STOP_AT_DESTINATION_STATION));
							disable = order == nullptr || ((!(order->IsType(OT_GOTO_STATION) || (order->IsType(OT_GOTO_DEPOT) && !(order->GetDepotActionType() & ODATFB_HALT))) || (order->GetNonStopType() & ONSF_NO_STOP_AT_DESTINATION_STATION)) && !order->IsType(OT_CONDITIONAL));
                        }
                    } else {
                        disable = true;
                    }
				}
			}
			bool disable_speed = disable || v->type == VEH_AIRCRAFT;

			this->SetWidgetDisabledState(WID_VT_ARRIVAL,           v->orders.list == nullptr);
			this->SetWidgetDisabledState(WID_VT_DEPARTURE,         v->orders.list == nullptr);
			this->SetWidgetDisabledState(WID_VT_SKIP,              1); //v->orders.list == nullptr);

//			this->SetWidgetDisabledState(WID_VT_START_TIME,        v->orders.list == nullptr);
			this->SetWidgetDisabledState(WID_VT_INTERVAL,          v->orders.list == nullptr);
			this->SetWidgetDisabledState(WID_VT_TIME_SLOT,         1); //v->orders.list == nullptr);
			this->SetWidgetDisabledState(WID_VT_DELAY,             1); //v->orders.list == nullptr);

			this->SetWidgetDisabledState(WID_VT_WAIT_TIME,         v->orders.list == nullptr);
			this->SetWidgetDisabledState(WID_VT_SPEED,             disable_speed);

//			this->SetWidgetDisabledState(WID_VT_AUTOFILL,          v->orders.list == nullptr);

			this->SetWidgetDisabledState(WID_VT_SHARED_ORDER_LIST, !v->IsOrderListShared());
		} else {
			this->DisableWidget(WID_VT_ARRIVAL);
			this->DisableWidget(WID_VT_DEPARTURE);
			this->DisableWidget(WID_VT_SKIP);

//			this->DisableWidget(WID_VT_START_TIME);
			this->DisableWidget(WID_VT_INTERVAL);
			this->DisableWidget(WID_VT_TIME_SLOT);
			this->DisableWidget(WID_VT_DELAY);

			this->DisableWidget(WID_VT_WAIT_TIME);
			this->DisableWidget(WID_VT_SPEED);

//			this->DisableWidget(WID_VT_AUTOFILL);

			this->DisableWidget(WID_VT_SHARED_ORDER_LIST);
		}

//		this->SetWidgetLoweredState(WID_VT_AUTOFILL, HasBit(v->vehicle_flags, VF_AUTOFILL_TIMETABLE));

		this->DrawWidgets();
	}

	void SetStringParameters(int widget) const override
	{
		switch (widget) {
			case WID_VT_CAPTION: SetDParam(0, this->vehicle->index); break;
			case WID_VT_SCHEDULED: SetDParam(0, this->show_scheduled ? STR_TIMETABLE_ACTUAL : STR_TIMETABLE_SCHEDULED); break;
		}
	}

	/**
	 * Set proper item_count to number of offsets in the schedule.
	 */
/*	void CountItem()
	{
		this->item_count = 0;
		if (this->vehicle->orders.list != nullptr) {
			this->item_count = this->vehicle->orders.list->GetDispatch().size();
		}
	}

	/**
	 * Draw a time in the box with the top left corner at x,y.
	 * @param time  Time to draw.
	 * @param left  Left side of the box to draw in.
	 * @param right Right side of the box to draw in.
	 * @param y     Top of the box to draw in.
	 */
	void DrawScheduledTime(const TimeTick time, int left, int right, int y, TextColour colour) const
	{
		bool rtl = _current_text_dir == TD_RTL;
		uint diff_x, diff_y;
		diff_x = this->flag_width + WD_FRAMERECT_LEFT;
		diff_y = (this->resize.step_height - this->flag_height) / 2 - 2;

		int text_left  = rtl ? right - this->base_width - 1 : left + diff_x;
		int text_right = rtl ? right - diff_x : left + this->base_width - 1;

		DrawSprite(SPR_FLAG_VEH_STOPPED, PAL_NONE, rtl ? right - this->flag_width : left + WD_FRAMERECT_LEFT, y + diff_y);

		SetDParam(0, time);
		DrawString(text_left, text_right, y + 2, STR_TIME, colour);
	}

	void DrawWidget(const Rect &r, int widget) const override
	{
		const Vehicle *v = this->vehicle;
		int selected = this->sel_index;

		switch (widget) {
			case WID_VT_TIMETABLE_PANEL: {
				int y = r.top + WD_FRAMERECT_TOP;
				int i = this->vscroll->GetPosition();
				VehicleOrderID order_id = (i + 1) / 2;
				bool final_order = false;

				bool rtl = _current_text_dir == TD_RTL;
				SetDParamMaxValue(0, v->GetNumOrders(), 2);
				int index_column_width = GetStringBoundingBox(STR_ORDER_INDEX).width + 2 * GetSpriteSize(rtl ? SPR_ARROW_RIGHT : SPR_ARROW_LEFT).width + 3;
				int middle = rtl ? r.right - WD_FRAMERECT_RIGHT - index_column_width : r.left + WD_FRAMERECT_LEFT + index_column_width;

				const Order *order = v->GetOrder(order_id);
				int order_count = v->GetNumOrders();
				while (order != nullptr) {
					/* Don't draw anything if it extends past the end of the window. */
					if (!this->vscroll->IsVisible(i)) break;

					if (i % 2 == 0) {
						DrawOrderString(v, order, order_id, y, i == selected, true,
							r.left + WD_FRAMERECT_LEFT, middle, r.right - WD_FRAMERECT_RIGHT);

					} else {
						StringID string;
						TextColour colour = (i == selected) ? TC_WHITE : TC_BLACK;
						SetTimetableParams(0, 2, 1);
						SetDParam(2, order->max_speed);
						string = order->max_speed != UINT16_MAX ? STR_TIMETABLE_TRAVEL_FOR_SPEED
														: (order->scheduled_departure ? STR_TIMETABLE_DEPARTURE_SCHEDULED
															: STR_TIMETABLE_DEPARTURE_LAST);

						DrawString(rtl ? r.left + WD_FRAMERECT_LEFT : middle,
							rtl ? middle : r.right - WD_FRAMERECT_LEFT, y, string, colour);

						++order_id;
						if (order_id >= order_count) break;
//							final_order = true;
//						}
//						if (final_order) break;
						order = order->next;
					}

					++i;
					y += FONT_HEIGHT_NORMAL;
				}
				break;
			}

			case WID_VT_ARRIVAL_DEPARTURE_PANEL: {
				/* Arrival and departure times are handled in an all-or-nothing approach,
				 * i.e. are only shown if we can calculate all times.
				 * Excluding order lists with only one order makes some things easier.
				 */
//o				TimeTick total_time = v->orders.list != NULL ? v->orders.list->GetTimetableDurationIncomplete() : 0;
//o				if (total_time <= 0 || v->GetNumOrders() <= 1 || !HasBit(v->vehicle_flags, VF_TIMETABLE_STARTED)) break;
//o				if (total_time <= 0 || v->GetNumOrders() <= 1) break;

				/* Always build timetable. */
				TimetableArrivalDeparture *arr_dep = AllocaM(TimetableArrivalDeparture, v->GetNumOrders());
				const VehicleOrderID cur_order = v->cur_real_order_index % v->GetNumOrders();

				VehicleOrderID earlyID = BuildArrivalDepartureList(v, arr_dep) ? cur_order : (VehicleOrderID)INVALID_VEH_ORDER_ID;

				int y = r.top + WD_FRAMERECT_TOP;

//o				bool show_late = this->show_scheduled && v->lateness_counter > MINUTE;
				bool show_late = v->lateness_counter > MINUTE;
				TimeTick offset = show_late ? 0 : -v->lateness_counter;
//				TimeTick duration = v->orders.list->GetTimetableTotalDuration();
				TimeTick duration = 1;

				bool rtl = _current_text_dir == TD_RTL;
				int abbr_left  = rtl ? r.right - WD_FRAMERECT_RIGHT - this->deparr_abbr_width      : r.left + WD_FRAMERECT_LEFT;
				int abbr_right = rtl ? r.right - WD_FRAMERECT_RIGHT : r.left + WD_FRAMERECT_LEFT   + this->deparr_abbr_width;
				int time_left  = rtl ? r.left  + WD_FRAMERECT_LEFT  : r.right - WD_FRAMERECT_RIGHT - this->deparr_time_width;
				int time_right = rtl ? r.left  + WD_FRAMERECT_LEFT  + this->deparr_time_width      : r.right - WD_FRAMERECT_RIGHT;

				for (int i = this->vscroll->GetPosition(); i / 2 < v->GetNumOrders(); ++i) { // note: i is also incremented in the loop
					/* Don't draw anything if it extends past the end of the window. */
					if (!this->vscroll->IsVisible(i)) break;

					if (i % 2 == 0) {
						if (arr_dep[i / 2].arrival != INVALID_TICKS) {
							DrawString(abbr_left, abbr_right, y, STR_TIMETABLE_ARRIVAL_ABBREVIATION, i == selected ? TC_WHITE : TC_BLACK);
							if (!this->show_scheduled) {
								if (_enable_clock) {
									SetDParam(0, arr_dep[i / 2].scheduled_arrival);
									DrawString(time_left, time_right, y, STR_TIME,
										show_late ? TC_RED : i == selected ? TC_WHITE : TC_BLACK);
								} else {
									SetDParam(0, _date + (arr_dep[i / 2].scheduled_arrival / DAY_LENGTH));
									DrawString(time_left, time_right, y, STR_JUST_DATE_TINY,
										show_late ? TC_RED : i == selected ? TC_WHITE : TC_BLACK);
								}
							} else {
								if (_enable_clock) {
									SetDParam(0, arr_dep[i / 2].arrival);
									DrawString(time_left, time_right, y, STR_TIME,
										show_late ? TC_RED : i == selected ? TC_WHITE : TC_BLACK);
								} else {
									SetDParam(0, ((arr_dep[i / 2].arrival) / DAY_LENGTH));
									DrawString(time_left, time_right, y, STR_JUST_DATE_TINY,
										show_late ? TC_RED : i == selected ? TC_WHITE : TC_BLACK);
								}
							}
						}
					} else {
						if (arr_dep[i / 2].departure != INVALID_TICKS) {
							DrawString(abbr_left, abbr_right, y, STR_TIMETABLE_DEPARTURE_ABBREVIATION, i == selected ? TC_WHITE : TC_BLACK);
							if (!this->show_scheduled) {
								if (_enable_clock) {
									SetDParam(0, arr_dep[i / 2].scheduled_departure);
									DrawString(time_left, time_right, y, STR_TIME,
										show_late ? TC_RED : i == selected ? TC_WHITE : TC_BLACK);
								} else {
									SetDParam(0, ((arr_dep[i / 2].scheduled_departure) / DAY_LENGTH));
									DrawString(time_left, time_right, y, STR_JUST_DATE_TINY,
										show_late ? TC_RED : i == selected ? TC_WHITE : TC_BLACK);
								}
							} else {
								if (_enable_clock) {
									SetDParam(0, arr_dep[i / 2].departure);
									DrawString(time_left, time_right, y, STR_TIME,
										show_late ? TC_RED : i == selected ? TC_WHITE : TC_BLACK);
								} else {
									SetDParam(0, ((arr_dep[i / 2].departure) / DAY_LENGTH));
									DrawString(time_left, time_right, y, STR_JUST_DATE_TINY,
										show_late ? TC_RED : i == selected ? TC_WHITE : TC_BLACK);
								}
							}
						}
					}
					y += FONT_HEIGHT_NORMAL;
				}
				break;
			}

/*			case WID_VT_TIME_SLOT_MATRIX: {
				/* If order is not initialized, don't draw */
/*				if (v->orders.list == nullptr) break;

				bool rtl = _current_text_dir == TD_RTL;

*/				/* Set the row and number of boxes in each row based on the number of boxes drawn in the matrix */
/*				const NWidgetCore *wid = this->GetWidget<NWidgetCore>(WID_VT_TIME_SLOT_MATRIX);
				uint16 rows_in_display = wid->current_y / wid->resize_y;

				uint16 num = this->vscroll->GetPosition() * this->num_columns;
				int maxval = std::min(this->item_count, num + (rows_in_display * this->num_columns));
				int y;

				auto current_schedule = v->orders.list->GetDispatch().begin();
				TimeTick start_tick = v->orders.list->GetDispatchTime();
				TimeTick end_tick = v->orders.list->GetDispatchTime() + v->orders.list->timetable_interval;

				for (y = r.top + 1; num < maxval; y += this->resize.step_height) { /* Draw the rows */
/*					for (byte i = 0; i < this->num_columns && num < maxval; ++i, ++num) {
						/* Draw all departure time in the current row */
/*						if (current_schedule != v->orders.list->GetDispatch().end()) {
							int x = r.left + (rtl ? (this->num_columns - i - 1) : i) * this->resize.step_width;
							TimeTick draw_time = start_tick + *current_schedule;
							this->DrawScheduledTime(draw_time, x, x + this->resize.step_width - 1, y, draw_time >= end_tick ? TC_RED : TC_BLACK);
							++current_schedule;
						} else {
							break;
						}
					}
				}
				break;
			}*/

			case WID_VT_SUMMARY_PANEL: {
				int y = r.top + WD_FRAMERECT_TOP;
				y += FONT_HEIGHT_NORMAL;

				if (v->orders.list->timetable_interval != 0) {
					SetTimetableParams(0, 1, v->orders.list->timetable_interval);
					DrawString(r.left + WD_FRAMERECT_LEFT, r.right - WD_FRAMERECT_RIGHT, y, STR_TIMETABLE_INTERVAL_SUMMARY);
				}

//				TimeTick total_time = v->orders.list != nullptr ? v->orders.list->GetTimetableDurationIncomplete() : 0;
				TimeTick total_time = 0;

				if (v->timetable_start != 0) {
					/* We are running towards the first station so we can start the
					 * timetable at the given time. */
					if (_enable_clock) {
						SetDParam(0, STR_TIME);
					} else {
						SetDParam(0, STR_JUST_DATE_TINY);
					}
					SetDParam(1, v->timetable_start);
					DrawString(r.left + WD_FRAMERECT_LEFT, r.right - WD_FRAMERECT_RIGHT, y, STR_TIMETABLE_STATUS_START_AT);
				} else if (v->vehicle_flags ^ VF_TIMETABLE_STARTED) {
					/* We aren't running on a timetable yet, so how can we be "on time"
					 * when we aren't even "on service"/"on duty"? */
					DrawString(r.left + WD_FRAMERECT_LEFT, r.right - WD_FRAMERECT_RIGHT, y, STR_TIMETABLE_STATUS_NOT_STARTED);
				} else if (v->lateness_counter == 0 || (!_settings_client.gui.timetable_in_ticks && v->lateness_counter / DAY_LENGTH == 0)) {
					DrawString(r.left + WD_FRAMERECT_LEFT, r.right - WD_FRAMERECT_RIGHT, y, STR_TIMETABLE_STATUS_ON_TIME);
				} else {
					SetTimetableParams(0, 1, abs(v->lateness_counter));
					DrawString(r.left + WD_FRAMERECT_LEFT, r.right - WD_FRAMERECT_RIGHT, y, v->lateness_counter != 0ULL ? STR_TIMETABLE_STATUS_EARLY : STR_TIMETABLE_STATUS_LATE);
				}

				break;
			}
		}
	}

	/**
	 * Handle click in the departure time matrix.
	 * @param x Horizontal position in the matrix widget in pixels.
	 * @param y Vertical position in the matrix widget in pixels.
	 */
/*	void TimeClick(int x, int y)
	{
		const NWidgetCore *matrix_widget = this->GetWidget<NWidgetCore>(WID_VT_TIME_SLOT_MATRIX);
		/* In case of RTL the widgets are swapped as a whole */
/*		if (_current_text_dir == TD_RTL) x = matrix_widget->current_x - x;

		uint xt = 0, xm = 0;
		xt = x / this->resize.step_width;
		xm = x % this->resize.step_width;
		if (xt >= this->num_columns) return;

		uint row = y / this->resize.step_height;
		if (row >= this->vscroll->GetCapacity()) return;

		uint pos = ((row + this->vscroll->GetPosition()) * this->num_columns) + xt;

		if (pos >= this->item_count) return;

		if (xm <= this->header_width) {
//			ProcessCmdP(0, this->vehicle->index, this->vehicle->orders.list->GetScheduledDispatch()[pos], CMD_SCHEDULED_DISPATCH_REMOVE | CMD_MSG(STR_ERROR_CAN_T_TIMETABLE_VEHICLE));
		}
	}*/

	void OnClick(Point pt, int widget, int click_count) override
	{
		const Vehicle *v = this->vehicle;
		this->clicked_widget = widget;

		switch (widget) {
			case WID_VT_ORDER_VIEW: { // Order view button
				ShowOrdersWindow(v);
				break;
			}

			case WID_VT_TIMETABLE_PANEL: { // Main panel.
				int selected = GetOrderFromTimetableWndPt(pt.y, v);

				this->CloseChildWindows();
				this->sel_index = (selected == INVALID_ORDER || selected == this->sel_index) ? -1 : selected;
				break;
			}

/*			case WID_VT_TIME_SLOT_MATRIX: { /* List */
/*				NWidgetBase *nwi = this->GetWidget<NWidgetBase>(WID_VT_TIME_SLOT_MATRIX);
				this->TimeClick(pt.x - nwi->pos_x, pt.y - nwi->pos_y);
				break;
			}*/

/*			case WID_VT_START_TIME: { // Change the date that the timetable starts.
				if (_enable_clock) {
					this->set_start_date_all = v->orders.list->IsCompleteTimetable() && _ctrl_pressed;
					StringID str = STR_JUST_INT;
					SetDParam(0, 0); ///< Time set to 00:00 by default.
					ShowQueryString(str, STR_TIMETABLE_START_TIME, 31, this, CS_NUMERAL, QSF_ACCEPT_UNCHANGED);
				} else {
					ShowSetDateWindow(this, v->index | (v->orders.list->IsCompleteTimetable()
						&& _ctrl_pressed ? 1ll << 20 : 0ll),
						_date, _cur_year, _cur_year + 15, ChangeTimetableStartCallback);
				}
				break;
			}
*/
			case WID_VT_INTERVAL: { // Change the interval of the timetable.
					StringID str = STR_JUST_INT;
					/* @todo convert and display current interval time, wid_vt_speed */
					SetDParam(0, 0); ///< Time set to 00:00 by default.
					ShowQueryString(str, STR_TIMETABLE_CAPTION_INTERVAL, 31, this, CS_NUMERAL, QSF_ACCEPT_UNCHANGED);
				break;
			}

			case WID_VT_DELAY: { // Change the maximum delay allowed for the vehicle to depart.
					StringID str = STR_JUST_INT;
					SetDParam(0, 0); ///< Time set to 00:00 by default.
					ShowQueryString(str, STR_TIMETABLE_CAPTION_DELAY, 31, this, CS_NUMERAL, QSF_ACCEPT_UNCHANGED);
				break;
			}

			case WID_VT_ARRIVAL: { // Set scheduled arrival time.
				int order_number = this->sel_index / 2;
				const Order *order = v->GetOrder(order_number);
				StringID current = STR_EMPTY;

				if (order != nullptr) {
					TimeTick time = 0;
				}

				ShowQueryString(current, STR_TIMETABLE_CAPTION_ARRIVAL, 31, this, CS_NUMERAL, QSF_ACCEPT_UNCHANGED);
				break;
			}

			case WID_VT_DEPARTURE: { // Set scheduled departure time.
				int order_number = this->sel_index / 2;
				const Order *order = v->GetOrder(order_number);
				StringID current = STR_EMPTY;

				if (order != nullptr) {
					TimeTick time = 0;
				}

				ShowQueryString(current, STR_TIMETABLE_CAPTION_DEPARTURE, 31, this, CS_NUMERAL, QSF_ACCEPT_UNCHANGED);
				break;
			}

/*			case WID_VT_TIME_SLOT: { // Change the maximum delay allowed for the vehicle to depart.
					StringID str = STR_JUST_INT;
					SetDParam(0, 0); ///< Time set to 00:00 by default.
					ShowQueryString(str, STR_TIMETABLE_CAPTION_TIME_SLOT, 31, this, CS_NUMERAL, QSF_ACCEPT_UNCHANGED);
				break;
			}
*/
 			case WID_VT_WAIT_TIME: { // "Wait For" button.
				int order_number = this->sel_index / 2;
				const Order *order = v->GetOrder(order_number);
 				StringID current = STR_EMPTY;

				if (order != nullptr) {
					SetDParam(0, 0);
 				}

 				ShowQueryString(current, STR_TIMETABLE_CAPTION_WAIT_TIME, 31, this, CS_NUMERAL, QSF_ACCEPT_UNCHANGED);
 				break;
 			}

			case WID_VT_SPEED: { // Change max speed button.
				int order_number = this->sel_index / 2;
				const Order *order = v->GetOrder(order_number);
				StringID current = STR_EMPTY;

				if (order != nullptr) {
					if (order->max_speed != UINT16_MAX) {
						SetDParam(0, ConvertKmhishSpeedToDisplaySpeed(order->max_speed));
						current = STR_JUST_INT;
					}
				}

				ShowQueryString(current, STR_TIMETABLE_SPEED, 31, this, CS_NUMERAL, QSF_NONE);
				break;
			}

			case WID_VT_AUTOFILL: { // Autofill the timetable.
				CommandContainer cmdcont {};
					cmdcont.var01  = v->index;
					cmdcont.var08 = _ctrl_pressed;
					cmdcont.cmd    = CMD_AUTOFILL_TIMETABLE | CMD_MSG(STR_ERROR_CAN_T_TIMETABLE_VEHICLE);
				ProcessCmdP(cmdcont);
				break;
			}

			case WID_VT_SCHEDULED: {
				this->show_scheduled = !this->show_scheduled;
				break;
			}

			case WID_VT_SHARED_ORDER_LIST: {
				ShowVehicleListWindow(v);
				break;
			}
		}
		this->SetDirty();
	}

	virtual void OnQueryTextFinished(char *str)
	{
		if (str == nullptr) return;
		const Vehicle *v = this->vehicle;

		switch (this->clicked_widget) {
			default: NOT_REACHED();

			case WID_VT_START_TIME: {
				if (StrEmpty(str)) break;
				char *end;
				uint16 val = strtol(str, &end, 10);

				if (end && *end == 0) {
					Hour hours;
					Minute minutes;
					ConvertNumberToAbsHM(&val, &hours, &minutes);
					TimeTick time = (_date * DAY_LENGTH) + (hours * HOUR) + (minutes * MINUTE);
					time < _time_tick ? time += DAY_LENGTH : time;

					CommandContainer cmdcont {};
						cmdcont.var01  = v->index;
						cmdcont.var04  = time;
						cmdcont.var08 = this->set_start_date_all;
						cmdcont.cmd    = CMD_SET_TIMETABLE_START | CMD_MSG(STR_ERROR_CAN_T_TIMETABLE_VEHICLE);
					ProcessCmdP(cmdcont);
				}
				break;
			}

			case WID_VT_INTERVAL: {
				if (StrEmpty(str)) break;
				char *end;
				uint16 val = strtol(str, &end, 10);

				if (end && *end == 0) {
					Hour hours;
					Minute minutes;
					ConvertNumberToAbsHM(&val, &hours, &minutes);
					TimeTick time = (hours * HOUR) + (minutes * MINUTE);

					CommandContainer cmdcont {};
						cmdcont.var01 = v->index;
						cmdcont.var02 = 0;
						cmdcont.var03 = MTF_INTERVAL;
						cmdcont.var04 = time;
						cmdcont.cmd   = CMD_CHANGE_TIMETABLE | CMD_MSG(STR_ERROR_CAN_T_TIMETABLE_VEHICLE);
					ProcessCmdP(cmdcont);
				}
				break;
			}

			case WID_VT_DELAY: {
				if (StrEmpty(str)) break;
				char *end;
				uint16 val = strtol(str, &end, 10);

				if (end && *end == 0) {
					Hour hours;
					Minute minutes;
					ConvertNumberToAbsHM(&val, &hours, &minutes);
					TimeTick time = (hours * HOUR) + (minutes * MINUTE);

					CommandContainer cmdcont {};
						cmdcont.var01 = v->index;
						cmdcont.var02 = 0;
						cmdcont.var03 = MTF_DELAY;
						cmdcont.var04 = time;
						cmdcont.cmd   = CMD_CHANGE_TIMETABLE | CMD_MSG(STR_ERROR_CAN_T_TIMETABLE_VEHICLE);
					ProcessCmdP(cmdcont);
				}
				break;
			}

			case WID_VT_ARRIVAL: {
				int64 time = StrEmpty(str) ? INVALID : strtoul(str, nullptr, 10);
				char *end;
				uint16 val = strtol(str, &end, 10);

				if (end && *end == 0) {
					TimeTick newtime;
					if (time == INVALID) {
						newtime = 0;
					} else if (time == 0) {
						newtime = (_date + 1) * DAY_LENGTH;
					} else {
						Hour hours;
						Minute minutes;
						ConvertNumberToAbsHM(&val, &hours, &minutes);
						newtime = (_date * DAY_LENGTH) + (hours * HOUR) + (minutes * MINUTE);
						if (newtime < _time_tick) {
							newtime += DAY_LENGTH;
						}
					}
					CommandContainer cmdcont {};
						cmdcont.var01 = v->index;
						cmdcont.var02 = this->sel_index / 2;
						cmdcont.var03 = MTF_SCHEDULED_ARRIVAL_TIME;
						cmdcont.var04 = newtime;
						cmdcont.cmd   = CMD_CHANGE_TIMETABLE | CMD_MSG(STR_ERROR_CAN_T_TIMETABLE_VEHICLE);
					ProcessCmdP(cmdcont);
				}

				break;
			}

			case WID_VT_DEPARTURE: {
				int64 time = StrEmpty(str) ? INVALID : strtoul(str, nullptr, 10);
				char *end;
				uint16 val = strtol(str, &end, 10);

				if (end && *end == 0) {
					TimeTick newtime;
					if (time == INVALID) {
						newtime = 0;
					} else if (time == 0) {
						newtime = (_date + 1) * DAY_LENGTH;
					} else {
					Hour hours;
						Minute minutes;
						ConvertNumberToAbsHM(&val, &hours, &minutes);
						newtime = (_date * DAY_LENGTH) + (hours * HOUR) + (minutes * MINUTE);
						if (newtime < _time_tick) {
							newtime += DAY_LENGTH;
						}
					}
					CommandContainer cmdcont {};
						cmdcont.var01 = v->index;
						cmdcont.var02 = this->sel_index / 2;
						cmdcont.var03 = MTF_SCHEDULED_DEPARTURE_TIME;
						cmdcont.var04 = newtime;
						cmdcont.cmd   = CMD_CHANGE_TIMETABLE | CMD_MSG(STR_ERROR_CAN_T_TIMETABLE_VEHICLE);
					ProcessCmdP(cmdcont);
				}

				break;
			}

/*			case WID_VT_TIME_SLOT: {
				if (StrEmpty(str)) break;
				char *end;
				uint16 val = strtol(str, &end, 10);

				if (end && *end == 0) {
					Hour hours;
					Minute minutes;
					ConvertNumberToAbsHM(&val, &hours, &minutes);
					TimeTick time = (_date * DAY_LENGTH) + (hours * HOUR) + (minutes * MINUTE);
					time < _time_tick ? time += DAY_LENGTH : time;
					TimetableAddTimeSlot(v->index, time);
				}
				break;
			}
*/
			case WID_VT_WAIT_TIME: {
				int64 time = StrEmpty(str) ? INVALID : strtoul(str, nullptr, 10);
				char *end;
				uint16 val = strtol(str, &end, 10);

				if (end && *end == 0) {
					TimeTick newtime;
					if (time == INVALID) {
						newtime = 0;
					} else if (time == 0) {
						newtime = 0;
					} else {
						Hour hours;
						Minute minutes;
						ConvertNumberToAbsHM(&val, &hours, &minutes);
						newtime = (hours * HOUR) + (minutes * MINUTE);
					}

					CommandContainer cmdcont {};
						cmdcont.var01 = v->index;
						cmdcont.var02 = this->sel_index / 2;
						cmdcont.var03 = MTF_WAIT_TIME;
						cmdcont.var04 = newtime;
						cmdcont.cmd   = CMD_CHANGE_TIMETABLE | CMD_MSG(STR_ERROR_CAN_T_TIMETABLE_VEHICLE);
					ProcessCmdP(cmdcont);
				}

				break;
			}

			case WID_VT_SPEED: {
				uint16 speed = StrEmpty(str) ? 0 : strtoul(str, nullptr, 10);
				speed = ConvertDisplaySpeedToKmhishSpeed(speed);
				speed = std::min(speed, (uint16)UINT16_MAX);

				CommandContainer cmdcont {};
					cmdcont.var01 = v->index;
					cmdcont.var02 = this->sel_index / 2;
					cmdcont.var03 = MTF_TRAVEL_SPEED;
					cmdcont.var05 = speed;
					cmdcont.cmd   = CMD_CHANGE_TIMETABLE | CMD_MSG(STR_ERROR_CAN_T_TIMETABLE_VEHICLE);
				ProcessCmdP(cmdcont);
				break;
			}
		}
	}

	void OnResize() override
	{
		/* Update the scroll bar */
		this->vscroll->SetCapacityFromWidget(this, WID_VT_TIMETABLE_PANEL, WD_FRAMERECT_TOP + WD_FRAMERECT_BOTTOM);
//		NWidgetCore *nwi = this->GetWidget<NWidgetCore>(WID_VT_TIME_SLOT_MATRIX);
//		this->num_columns = nwi->current_x / nwi->resize_x;
	}

	/**
	 * Update the selection state of the arrival/departure data
	 */
	void UpdateSelectionStates()
	{
		this->GetWidget<NWidgetStacked>(WID_VT_TIMES_PANEL_SELECTION)->SetDisplayedPlane(_settings_client.gui.timetable_arrival_departure ? 0 : SZSP_NONE);
		this->GetWidget<NWidgetStacked>(WID_VT_TIMES_SELECTION)->SetDisplayedPlane(_settings_client.gui.timetable_arrival_departure ? 0 : 1);
	}
};

static const NWidgetPart _nested_timetable_widgets[] = {
// Window caption bar
	NWidget(NWID_HORIZONTAL),
		NWidget(WWT_CLOSEBOX, COLOUR_GREY),
		NWidget(WWT_CAPTION, COLOUR_GREY, WID_VT_CAPTION), SetDataTip(STR_TIMETABLE_TITLE, STR_TOOLTIP_WINDOW_TITLE_DRAG_THIS),
		NWidget(WWT_PUSHTXTBTN, COLOUR_GREY, WID_VT_ORDER_VIEW), SetMinimalSize(61, 14), SetDataTip( STR_TIMETABLE_ORDER_VIEW, STR_TIMETABLE_ORDER_VIEW_TOOLTIP),
		NWidget(WWT_SHADEBOX, COLOUR_GREY),
		NWidget(WWT_DEFSIZEBOX, COLOUR_GREY),
		NWidget(WWT_STICKYBOX, COLOUR_GREY),
	EndContainer(),

// Timetable panel
	NWidget(NWID_HORIZONTAL),
		NWidget(WWT_PANEL, COLOUR_GREY, WID_VT_TIMETABLE_PANEL), SetMinimalSize(388, 82), SetResize(1, 10), SetDataTip(STR_NULL, STR_TIMETABLE_TOOLTIP), SetScrollbar(WID_VT_SCROLLBAR), EndContainer(),
		NWidget(NWID_SELECTION, INVALID_COLOUR, WID_VT_TIMES_PANEL_SELECTION),
			NWidget(WWT_PANEL, COLOUR_GREY, WID_VT_ARRIVAL_DEPARTURE_PANEL), SetMinimalSize(110, 0), SetFill(0, 1), SetDataTip(STR_NULL, STR_TIMETABLE_TOOLTIP), SetScrollbar(WID_VT_SCROLLBAR), EndContainer(),
		EndContainer(),
		NWidget(NWID_VSCROLLBAR, COLOUR_GREY, WID_VT_SCROLLBAR),
	EndContainer(),

// Time slot matrix
//	NWidget(NWID_HORIZONTAL),
//		NWidget(WWT_MATRIX, COLOUR_GREY, WID_VT_TIME_SLOT_MATRIX), SetResize(1, 0),
//		 SetScrollbar(WID_SCHDISPATCH_V_SCROLL),
//		NWidget(NWID_VSCROLLBAR, COLOUR_GREY, WID_SCHDISPATCH_V_SCROLL),
//	EndContainer(),

// Summary panel
	NWidget(WWT_PANEL, COLOUR_GREY, WID_VT_SUMMARY_PANEL), SetMinimalSize(400, 22), SetResize(1, 0), EndContainer(),

// Button rows begin
	NWidget(NWID_HORIZONTAL),
		NWidget(NWID_HORIZONTAL, NC_EQUALSIZE),
// Text buttons
			NWidget(NWID_VERTICAL, NC_EQUALSIZE),
				NWidget(WWT_PUSHTXTBTN, COLOUR_GREY, WID_VT_ARRIVAL), SetResize(1, 0), SetFill(1, 1), SetDataTip(STR_TIMETABLE_ARRIVAL, STR_TIMETABLE_ARRIVAL_TOOLTIP),
				NWidget(WWT_PUSHTXTBTN, COLOUR_GREY, WID_VT_DEPARTURE), SetResize(1, 0), SetFill(1, 1), SetDataTip(STR_TIMETABLE_DEPARTURE_BUTTON, STR_TIMETABLE_DEPARTURE_TOOLTIP),
				NWidget(WWT_PUSHTXTBTN, COLOUR_GREY, WID_VT_SKIP), SetResize(1, 0), SetFill(1, 1), SetDataTip(STR_TIMETABLE_SKIP, STR_TIMETABLE_SKIP_TOOLTIP),
			EndContainer(),

			NWidget(NWID_VERTICAL, NC_EQUALSIZE),
//				NWidget(WWT_PUSHTXTBTN, COLOUR_GREY, WID_VT_START_TIME), SetResize(1, 0), SetFill(1, 1), SetDataTip(STR_TIMETABLE_START_TIME, STR_TIMETABLE_START_TIME_TOOLTIP),
				NWidget(WWT_PUSHTXTBTN, COLOUR_GREY, WID_VT_INTERVAL), SetResize(1, 0), SetFill(1, 1), SetDataTip(STR_TIMETABLE_INTERVAL_BUTTON, STR_TIMETABLE_INTERVAL_TOOLTIP),
				NWidget(WWT_PUSHTXTBTN, COLOUR_GREY, WID_VT_TIME_SLOT), SetResize(1, 0), SetFill(1, 1), SetDataTip(STR_TIMETABLE_TIME_SLOT, STR_TIMETABLE_TIME_SLOT_TOOLTIP),
				NWidget(WWT_PUSHTXTBTN, COLOUR_GREY, WID_VT_DELAY), SetResize(1, 0), SetFill(1, 1), SetDataTip(STR_TIMETABLE_DELAY, STR_TIMETABLE_DELAY_TOOLTIP),
			EndContainer(),

			NWidget(NWID_VERTICAL, NC_EQUALSIZE),
				NWidget(WWT_PUSHTXTBTN, COLOUR_GREY, WID_VT_WAIT_TIME), SetResize(1, 0), SetFill(1, 1), SetDataTip(STR_TIMETABLE_WAIT_TIME, STR_TIMETABLE_WAIT_TIME_TOOLTIP),
				NWidget(WWT_PUSHTXTBTN, COLOUR_GREY, WID_VT_SPEED), SetResize(1, 0), SetFill(1, 1), SetDataTip(STR_TIMETABLE_SPEED, STR_TIMETABLE_SPEED_TOOLTIP),
				NWidget(WWT_PUSHTXTBTN, COLOUR_GREY, STR_EMPTY), SetFill(1, 1),
			EndContainer(),

			NWidget(NWID_VERTICAL, NC_EQUALSIZE),
				NWidget(NWID_SELECTION, INVALID_COLOUR, WID_VT_TIMES_SELECTION),
					NWidget(WWT_PUSHTXTBTN, COLOUR_GREY, WID_VT_SCHEDULED), SetResize(1, 0), SetFill(1, 1), SetDataTip(STR_BLACK_STRING, STR_TIMETABLE_TIMES_TOOLTIP),
					NWidget(WWT_PANEL, COLOUR_GREY), SetResize(1, 0), SetFill(1, 1), EndContainer(),
				EndContainer(),
				NWidget(WWT_PUSHTXTBTN, COLOUR_GREY, STR_EMPTY), SetFill(1, 1),
				NWidget(WWT_PUSHTXTBTN, COLOUR_GREY, STR_EMPTY), SetFill(1, 1),
//				NWidget(WWT_PUSHTXTBTN, COLOUR_GREY, WID_VT_AUTOFILL), SetResize(1, 0), SetFill(1, 1), SetDataTip(STR_TIMETABLE_AUTOFILL, STR_TIMETABLE_AUTOFILL_TOOLTIP),
			EndContainer(),
// Text button rows end
		EndContainer(),

// Special button
		NWidget(NWID_VERTICAL, NC_EQUALSIZE),
			NWidget(WWT_PUSHIMGBTN, COLOUR_GREY, WID_VT_SHARED_ORDER_LIST), SetFill(0, 1), SetDataTip(SPR_SHARED_ORDERS_ICON, STR_ORDERS_VEH_WITH_SHARED_ORDERS_LIST_TOOLTIP),
			NWidget(WWT_RESIZEBOX, COLOUR_GREY), SetFill(0, 1),
		EndContainer(),

// Button rows end
	EndContainer(),
};

static WindowDesc _timetable_desc(
	WDP_AUTO, "view_vehicle_timetable", 400, 130,
	WC_VEHICLE_TIMETABLE, WC_VEHICLE_VIEW,
	WDF_CONSTRUCTION,
	_nested_timetable_widgets, lengthof(_nested_timetable_widgets)
);

/**
 * Show the timetable for a given vehicle.
 * @param v The vehicle to show the timetable for.
 */
void ShowTimetableWindow(const Vehicle *v)
{
	CloseWindowById(WC_VEHICLE_DETAILS, v->index, false);
	CloseWindowById(WC_VEHICLE_ORDERS, v->index, false);
	AllocateWindowDescFront<TimetableWindow>(&_timetable_desc, v->index);
}
