/*
 * This file is part of OpenTTD.
 * OpenTTD is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, version 2.
 * OpenTTD is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details. You should have received a copy of the GNU General Public License along with OpenTTD. If not, see <http://www.gnu.org/licenses/>.
 */

/** @file  vehicle_base.h Base class for all vehicles. */



#ifndef VEHICLE_BASE_H
#define VEHICLE_BASE_H



#include <list>
#include <map>

#include "vehicle_type.h"

#include "core/enum_type.hpp"
#include "core/pool_type.hpp"
#include "core/smallmap_type.hpp"
#include "group_type.h"
#include "track_type.h"
#include "command_type.h"
#include "order_type.h"
#include "order_base.h"
#include "cargopacket.h"
#include "texteff.hpp"
#include "engine_type.h"
#include "order_func.h"
#include "transport_type.h"
#include "base_consist.h"
#include "network/network.h"
#include "saveload/saveload.h"
#include "timetable.h"



/** Vehicle status bits in #Vehicle::vehstatus, e.g. vehicle "is". */
enum VehStatus {
	VS_HIDDEN          = 1 <<  0, ///< Vehicle is not visible.
	VS_STOPPED         = 1 <<  1, ///< Vehicle is stopped by the player.
	VS_UNCLICKABLE     = 1 <<  2, ///< Vehicle is not clickable by the user (shadow vehicles).
	VS_DEFPAL          = 1 <<  3, ///< Use default vehicle palette. @see DoDrawVehicle
	VS_TRAIN_SLOWING   = 1 <<  4, ///< Train is slowing down.
	VS_SHADOW          = 1 <<  5, ///< Vehicle is a shadow vehicle.
	VS_AIRCRAFT_BROKEN = 1 <<  6, ///< Aircraft is broken down.
	VS_CRASHED         = 1 <<  7, ///< Vehicle is crashed.
	VS_WAITING         = 1 <<  8, ///< Vehicle is waiting.
	VS_REVERSING       = 1 <<  9, ///< Vehicle is turning around.
	VS_ENTER_STATION   = 1 << 10, ///< Vehicle is entering a station.
	VS_AT_STATION      = 1 << 11, ///< Vehicle is at a station.
	VS_LEAVE_STATION   = 1 << 12, ///< Vehicle is leaving a station.
	VS_ENTER_WAYPOINT  = 1 << 13, ///< Vehicle is entering a waypoint.
	VS_AT_WAYPOINT     = 1 << 14, ///< Vehicle is at a waypoint.
	VS_LEAVE_WAYPOINT  = 1 << 15, ///< Vehicle is leaving a waypoint.
	VS_ENTER_DEPOT     = 1 << 16, ///< Vehicle is entering a depot.
	VS_AT_DEPOT        = 1 << 17, ///< Vehicle is at a depot.
	VS_LEAVE_DEPOT     = 1 << 18, ///< Vehicle is leaving a depot.
	VS_LOADING         = 1 << 19, ///< Vehicle is loading.
	VS_UNLOADING       = 1 << 20, ///< Vehicle is unloading.
	VS_TRAVELLING      = 1 << 21, ///< Vehicle is travelling.
	VS_HALTED          = 1 << 22, ///< Vehicle is temporarily stopped.
	VS_DEFECT          = 1 << 23, ///< Vehicle is defective.
	VS_SERVICING       = 1 << 24, ///< Vehicle is in service.
	VS_PARKED          = 1 << 25, ///< Vehicle is parked.
	VS_STATION_LISTED  = 1 << 26, ///< Vehicle is on station loading list.
	VS_AT_BAY          = 1 << 27,
	VS_AT_BAY1         = 1 << 28,
	VS_AT_BAY2         = 1 << 29,
};



/** Bit numbers in #Vehicle::vehicle_flags, e.g. vehicle "has". */
enum VehicleFlags {
	VF_LOADING_FINISHED        =       0, ///< Vehicle has finished loading.
	VF_CARGO_UNLOADING         = 1 <<  0, ///< Vehicle is unloading cargo.
	VF_BUILT_AS_PROTOTYPE      = 1 <<  1, ///< Vehicle is a prototype (accepted as exclusive preview).
	VF_TIMETABLE_STARTED       = 1 <<  2, ///< Whether the vehicle has started running on the timetable yet.
	VF_AUTOFILL_TIMETABLE      = 1 <<  3, ///< Whether the vehicle should fill in the timetable automatically.
	VF_AUTOFILL_PRES_WAIT_TIME = 1 <<  4, ///< Whether non-destructive auto-fill should preserve waiting times
	VF_STOP_LOADING            = 1 <<  5, ///< Don't load anymore during the next load cycle.
	VF_PATHFINDER_LOST         = 1 <<  6, ///< Vehicle's pathfinder is lost.
	VF_SERVINT_IS_CUSTOM       = 1 <<  7, ///< Service interval is custom.
	VF_SERVINT_IS_PERCENT      = 1 <<  8, ///< Service interval is percent.
	VF_SHARED_ORDERS           = 1 <<  9, ///< Vehicle shares orders with other vehicle(s).
	VF_COULD_REVERSE           = 1 << 10, ///< Vehicle could reverse at this point.
};



/** Bit numbers used to indicate which of the #NewGRFCache values are valid. */
enum NewGRFCacheValidValues {
	NCVV_POSITION_CONSIST_LENGTH   = 0, ///< This bit will be set if the NewGRF var 40 currently stored is valid.
	NCVV_POSITION_SAME_ID_LENGTH   = 1, ///< This bit will be set if the NewGRF var 41 currently stored is valid.
	NCVV_CONSIST_CARGO_INFORMATION = 2, ///< This bit will be set if the NewGRF var 42 currently stored is valid.
	NCVV_COMPANY_INFORMATION       = 3, ///< This bit will be set if the NewGRF var 43 currently stored is valid.
	NCVV_POSITION_IN_VEHICLE       = 4, ///< This bit will be set if the NewGRF var 4D currently stored is valid.
	NCVV_END,                           ///< End of the bits.
};



/** Cached often queried (NewGRF) values */
struct NewGRFCache {
	/* Values calculated when they are requested for the first time after invalidating the NewGRF cache. */
	uint32 position_consist_length;   ///< Cache for NewGRF var 40.
	uint32 position_same_id_length;   ///< Cache for NewGRF var 41.
	uint32 consist_cargo_information; ///< Cache for NewGRF var 42. (Note: The cargotype is untranslated in the cache
									  ///<  because the accessing GRF is yet unknown.)
	uint32 company_information;       ///< Cache for NewGRF var 43.
	uint32 position_in_vehicle;       ///< Cache for NewGRF var 4D.
	uint8  cache_valid;               ///< Bitset that indicates which cache values are valid.
};



/** Meaning of the various bits of the visual effect. */
enum VisualEffect {
	VE_OFFSET_START        = 0,                 ///< First bit that contains the offset (0 = front, 8 = centre, 15 = rear)
	VE_OFFSET_COUNT        = 4,                 ///< Number of bits used for the offset
	VE_OFFSET_CENTRE       = 8,                 ///< Value of offset corresponding to a position above the centre of the vehicle

	VE_TYPE_START          = 4,                 ///< First bit used for the type of effect
	VE_TYPE_COUNT          = 2,                 ///< Number of bits used for the effect type
	VE_TYPE_DEFAULT        = 0,                 ///< Use default from engine class
	VE_TYPE_STEAM          = 1,                 ///< Steam plumes
	VE_TYPE_DIESEL         = 2,                 ///< Diesel fumes
	VE_TYPE_ELECTRIC       = 3,                 ///< Electric sparks

	VE_DISABLE_EFFECT      = 6,                 ///< Flag to disable visual effect
	VE_ADVANCED_EFFECT     = VE_DISABLE_EFFECT, ///< Flag for advanced effects
	VE_DISABLE_WAGON_POWER = 7,                 ///< Flag to disable wagon power

	VE_DEFAULT = 0xFF,          ///< Default value to indicate that visual effect should be based on engine class
};



/** Models for spawning visual effects. */
enum VisualEffectSpawnModel {
	VESM_NONE              = 0, ///< No visual effect
	VESM_STEAM,                 ///< Steam model
	VESM_DIESEL,                ///< Diesel model
	VESM_ELECTRIC,              ///< Electric model

	VESM_END
};



/**
 * Enum to handle ground vehicle subtypes.
 * This is defined here instead of at #GroundVehicle because some common function require access to these flags.
 * Do not access it directly unless you have to. Use the subtype access functions.
 */
enum GroundVehicleSubtypeFlags {
	GVSF_FRONT            = 0, ///< Leading engine of a consist.
	GVSF_ARTICULATED_PART = 1, ///< Articulated part of an engine.
	GVSF_WAGON            = 2, ///< Wagon (not used for road vehicles).
	GVSF_ENGINE           = 3, ///< Engine that can be front engine, but might be placed behind another engine
							   ///< (not used for road vehicles).
	GVSF_FREE_WAGON       = 4, ///< First in a wagon chain (in depot) (not used for road vehicles).
	GVSF_MULTIHEADED      = 5, ///< Engine is multiheaded (not used for road vehicles).
};



/**
 * Iterator to iterate orders
 * Supports deletion of current order
 */
struct OrderIterator {
	typedef Order value_type;
	typedef Order* pointer;
	typedef Order& reference;
	typedef size_t difference_type;
	typedef std::forward_iterator_tag iterator_category;

	explicit OrderIterator(OrderList *list) : list(list), prev(nullptr)
	{
		this->order = (this->list == nullptr) ? nullptr : this->list->GetFirstOrder();
	}

	bool operator==(const OrderIterator &other) const { return this->order == other.order; }
	bool operator!=(const OrderIterator &other) const { return !(*this == other); }
	Order * operator*() const { return this->order; }
	OrderIterator & operator++()
	{
		this->prev = (this->prev == nullptr) ? this->list->GetFirstOrder() : this->prev->next;
		this->order = (this->prev == nullptr) ? nullptr : this->prev->next;
		return *this;
	}

private:
	OrderList *list;
	Order *order;
	Order *prev;
};



/**
 * Iterable ensemble of orders
 */
struct IterateWrapper {
	OrderList *list;
	IterateWrapper(OrderList *list = nullptr) : list(list) {}
	OrderIterator begin() { return OrderIterator(this->list); }
	OrderIterator end() { return OrderIterator(nullptr); }
	bool empty() { return this->begin() == this->end(); }
};



/** Cached often queried values common to all vehicles. */
struct VehicleCache {
	uint16 cached_max_speed;          ///< Maximum speed of the consist
									  ///< (minimum of the max speed of all vehicles in the consist).
	TimeTick cached_cargo_age_period; ///< Number of ticks before carried cargo is aged.
	byte cached_vis_effect;  		  ///< Visual effect to show (see #VisualEffect)
};



/** Sprite sequence for a vehicle part. */
struct VehicleSpriteSeq {
	PalSpriteID seq[4];
	uint count;

	bool operator==(const VehicleSpriteSeq &other) const
	{
		return this->count == other.count && MemCmpT<PalSpriteID>(this->seq, other.seq, this->count) == 0;
	}

	bool operator!=(const VehicleSpriteSeq &other) const
	{
		return !this->operator==(other);
	}

	/**
	 * Check whether the sequence contains any sprites.
	 */
	bool IsValid() const
	{
		return this->count != 0;
	}

	/**
	 * Clear all information.
	 */
	void Clear()
	{
		this->count = 0;
	}

	/**
	 * Assign a single sprite to the sequence.
	 */
	void Set(SpriteID sprite)
	{
		this->count = 1;
		this->seq[0].sprite = sprite;
		this->seq[0].pal = 0;
	}

	/**
	 * Copy data from another sprite sequence, while dropping all recolouring information.
	 */
	void CopyWithoutPalette(const VehicleSpriteSeq &src)
	{
		this->count = src.count;
		for (uint i = 0; i < src.count; ++i) {
			this->seq[i].sprite = src.seq[i].sprite;
			this->seq[i].pal = 0;
		}
	}

	void GetBounds(Rect *bounds) const;
	void Draw(int x, int y, PaletteID default_pal, bool force_pal) const;
};



/**
 * Cache for vehicle sprites and values relating to whether they should be updated before drawing,
 * or calculating the viewport.
 */
struct MutableSpriteCache {
	Direction last_direction;     ///< Last direction we obtained sprites for
	bool revalidate_before_draw;  ///< We need to do a GetImage() and check bounds before drawing this sprite
	Rect old_coord;               ///< Co-ordinates from the last valid bounding box
	bool is_viewport_candidate;   ///< This vehicle can potentially be drawn on a viewport
	VehicleSpriteSeq sprite_seq;  ///< Vehicle appearance.
};



/** A vehicle pool for a little over 1 million vehicles. */
typedef Pool<Vehicle, VehicleID, 512, 0xFF000> VehiclePool;
extern VehiclePool _vehicle_pool;



/* Some declarations of functions, so we can make them friendly */
struct GroundVehicleCache;
struct LoadgameState;
extern bool LoadOldVehicle(LoadgameState *ls, int num);
extern void FixOldVehicles();

struct GRFFile;



/**
 * Simulated cargo type and capacity for prediction of future links.
 */
struct RefitDesc {
	CargoID cargo;    ///< Cargo type the vehicle will be carrying.
	uint16 capacity;  ///< Capacity the vehicle will have.
	uint16 remaining; ///< Capacity remaining from before the previous refit.
	RefitDesc(CargoID cargo, uint16 capacity, uint16 remaining) :
			cargo(cargo), capacity(capacity), remaining(remaining) {}
};



/** %Vehicle data structure. */
struct Vehicle : VehiclePool::PoolItem<&_vehicle_pool>, BaseVehicle, BaseConsist {
//private:
//public:
	typedef std::list<RefitDesc> RefitList;
	typedef std::map<CargoID, uint> CapacitiesMap;

	Vehicle *next;                      ///< pointer to the next vehicle in the chain
	Vehicle *previous;                  ///< NOSAVE: pointer to the previous vehicle in the chain
	Vehicle *first;                     ///< NOSAVE: pointer to the first vehicle in the chain

	Vehicle *next_shared;               ///< pointer to the next vehicle that shares the order
	Vehicle *previous_shared;           ///< NOSAVE: pointer to the previous vehicle in the shared order chain

	friend void FixOldVehicles();
	friend void AfterLoadVehicles(bool part_of_load);             ///< So we can set the #previous and #first pointers while loading
	friend bool LoadOldVehicle(LoadgameState *ls, int num);       ///< So we can set the proper next pointer while loading
	/* So we can use private/protected variables in the saveload code */
	friend class SlVehicleCommon;
	friend class SlVehicleDisaster;
	friend void Ptrs_VEHS();

	TileIndex tile;                     ///< Current tile index

	/**
	 * Heading for this tile.
	 * For airports and train stations this tile does not necessarily belong to the destination station,
	 * but it can be used for heuristic purposes to estimate the distance.
	 */
	TileIndex dest_tile;

	Money profit_this_year;             ///< Profit this year << 8, low 8 bits are fract
	Money profit_last_year;             ///< Profit last year << 8, low 8 bits are fract
	Money value;                        ///< Value of the vehicle

	CargoPayment *cargo_payment;        ///< The cargo payment we're currently in

	mutable Rect coord;                 ///< NOSAVE: Graphical bounding box of the vehicle, i.e. what to redraw on moves.

	Vehicle *hash_viewport_next;        ///< NOSAVE: Next vehicle in the visual location hash.
	Vehicle **hash_viewport_prev;       ///< NOSAVE: Previous vehicle in the visual location hash.

	Vehicle *hash_tile_next;            ///< NOSAVE: Next vehicle in the tile location hash.
	Vehicle **hash_tile_prev;           ///< NOSAVE: Previous vehicle in the tile location hash.
	Vehicle **hash_tile_current;        ///< NOSAVE: Cache of the current hash chain.

	SpriteID colourmap;                 ///< NOSAVE: cached colour mapping

	/* Related to age and service time */
	Year build_year;                    ///< Year the vehicle has been built.
	Date age;                           ///< Age in days
	Date max_age;                       ///< Maximum age
	Date date_of_last_service;          ///< Last date the vehicle had a service at a depot.
	uint16 reliability;                 ///< Reliability.
	uint16 reliability_spd_dec;         ///< Reliability decrease speed.
	TimeTick service_time;              ///< Time needed for servicing the vehicle.
	byte breakdown_ctr;                 ///< Counter for managing breakdown events. @see Vehicle::HandleBreakdown
	byte breakdown_delay;               ///< Counter for managing breakdown length.
	byte breakdowns_since_last_service; ///< Counter for the amount of breakdowns.
	byte breakdown_chance;              ///< Current chance of breakdowns.

	int32 x_pos;                        ///< x coordinate.
	int32 y_pos;                        ///< y coordinate.
	int32 z_pos;                        ///< z coordinate.
	Direction direction;                ///< facing

	Owner owner;                        ///< Which company owns the vehicle?
	/**
	 * currently displayed sprite index
	 * 0xfd == custom sprite, 0xfe == custom second head sprite
	 * 0xff == reserved for another custom sprite
	 */
	byte spritenum;
	byte x_extent;                      ///< x-extent of vehicle bounding box
	byte y_extent;                      ///< y-extent of vehicle bounding box
	byte z_extent;                      ///< z-extent of vehicle bounding box
	int8 x_bb_offs;                     ///< x offset of vehicle bounding box
	int8 y_bb_offs;                     ///< y offset of vehicle bounding box
	int8 x_offs;                        ///< x offset for vehicle sprite
	int8 y_offs;                        ///< y offset for vehicle sprite
	EngineID engine_type;               ///< The type of engine used for this vehicle.

	TextEffectID fill_percent_te_id;    ///< a text-effect id to a loading indicator object
	UnitID unitnumber;                  ///< unit number, for display purposes only

	uint16 cur_speed;                   ///< current speed
	byte subspeed;                      ///< fractional speed
	byte acceleration;                  ///< used by train & aircraft
	uint32 motion_counter;              ///< counter to occasionally play a vehicle sound.
	byte progress;                      ///< The percentage (if divided by 256) this vehicle already crossed the tile unit.

	byte random_bits;                   ///< Bits used for determining which randomized variational spritegroups to use when drawing.
	byte waiting_triggers;              ///< Triggers to be yet matched before rerandomizing the random bits.

	StationID last_station_visited;     ///< The last station we stopped at.
	StationID last_loading_station;     ///< Last station the vehicle has stopped at and could possibly leave from with any cargo loaded.

	CargoID cargo_type;                 ///< type of cargo this vehicle is carrying
	byte cargo_subtype;                 ///< Used for livery refits (NewGRF variations)
	uint16 cargo_cap;                   ///< total capacity
	uint16 refit_cap;                   ///< Capacity left over from before last refit.
	VehicleCargoList cargo;             ///< The cargo this vehicle is carrying
	TimeTick cargo_age_counter;         ///< Ticks till cargo is aged next.
	int8 trip_occupancy;                ///< NOSAVE: Occupancy of vehicle of the current trip (updated after leaving a station).

	byte day_counter;                   ///< Increased by one for each day
	TimeTick tick_counter;              ///< Increased by one for each tick
	TimeTick running_ticks;             ///< Number of ticks this vehicle was not stopped this day

	int64 vehstatus;                    ///< Status
	Order current_order;                ///< The current order (+ status, like: loading)

	union {
		OrderList *list;                ///< Pointer to the order list for this vehicle
		Order     *old;                 ///< Only used during conversion of old save games
	} orders;                           ///< The orders currently assigned to the vehicle.

	TimeTick load_unload_ticks;         ///< Ticks to wait before starting next cycle.

	GroupID group_id;                   ///< Index of group Pool array
	byte subtype;                       ///< subtype (Filled with values from #AircraftSubType/#DisasterSubType/#EffectVehicleType
										///<									/#GroundVehicleSubtypeFlags)

	NewGRFCache grf_cache;              ///< Cache of often used calculated NewGRF values
	VehicleCache vcache;                ///< Cache of often used vehicle values.

	mutable MutableSpriteCache sprite_cache; ///< Cache of sprites and values related to recalculating them, see #MutableSpriteCache

	Vehicle(VehicleType type = VEH_INVALID);

	void PreDestructor();
	/** We want to 'destruct' the right class. */
	virtual ~Vehicle();

	void CancelReservation(StationID next, Station *st);

	GroundVehicleCache *GetGroundVehicleCache();
	const GroundVehicleCache *GetGroundVehicleCache() const;

	uint16 &GetGroundVehicleFlags();
	const uint16 &GetGroundVehicleFlags() const;

	bool HandleDeparture();

	void GetConsistFreeCapacities(SmallMap<CargoID, uint> &capacities) const;

	uint GetConsistTotalCapacity() const;

	virtual void MarkDirty();
	virtual void UpdateDeltaXY();
	uint GetOldAdvanceSpeed(uint speed);
	static uint GetAdvanceSpeed(uint speed);
	uint GetAdvanceDistance();
	virtual ExpensesType GetExpenseType(bool income) const;
	virtual void PlayLeaveStationSound() const;
	virtual bool IsPrimaryVehicle() const;

	const Engine *GetEngine() const;

	virtual void GetImage(Direction direction, EngineImageType image_type, VehicleSpriteSeq *result) const;

	const GRFFile *GetGRF() const;
	uint32 GetGRFID() const;

	void InvalidateNewGRFCache();
	void InvalidateNewGRFCacheOfChain();
	bool IsGroundVehicle() const;
	virtual int GetDisplaySpeed() const;
	virtual int GetDisplayMaxSpeed() const;
	virtual int GetCurrentMaxSpeed();
	virtual Money GetRunningCost() const;
	virtual bool IsInDepot() const;
	virtual bool IsChainInDepot() const;
	bool IsStoppedInDepot() const;
	virtual bool Tick();
	virtual void OnNewDay() {};

	void ShiftDates(int interval);

	virtual uint Crash(bool flooded);
	virtual Trackdir GetVehicleTrackdir() const;
	Money GetDisplayRunningCost() const;
	Money GetDisplayProfitThisYear() const;
	Money GetDisplayProfitLastYear() const;

	void SetNext(Vehicle *next);

	Vehicle *Next() const;
	Vehicle *Previous() const;
	Vehicle *First() const;
	Vehicle *Last();
	const Vehicle *Last() const;
	Vehicle *Move(int n);
	const Vehicle *Move(int n) const;
	Order *GetFirstOrder() const;
	Vehicle *NextShared() const;
	Vehicle *PreviousShared() const;
	Vehicle *FirstShared() const;
	bool IsOrderListShared() const;
	VehicleOrderID GetNumOrders() const;
	StationIDStack GetNextStoppingStation() const;
	void ResetRefitCaps();
	void CopyVehicleConfigAndStatistics(const Vehicle *src);

	bool HandleBreakdown();

	bool NeedsAutorenewing(const Company *c, bool use_renew_setting = true) const;

	bool NeedsServicing() const;
	bool NeedsAutomaticServicing() const;

	virtual TileIndex GetOrderStationLocation(StationID station);
	virtual bool FindClosestDepot(TileIndex *location, DestinationID *destination, bool *reverse);

	virtual void SetDestTile(TileIndex tile);

	CommandCost SendToDepot(ProcessCmdFlag flags, DepotCommand command);

	void UpdateVisualEffect(bool allow_power_change);
	void ShowVisualEffect();

	void UpdatePosition();
	void UpdateViewport(bool dirty);
	void UpdateBoundingBoxCoordinates(bool update_cache) const;
	void UpdatePositionAndViewport();
	bool MarkAllViewportsDirty() const;

	uint16 GetServiceInterval() const;

	void SetServiceInterval(uint16 interval);

	bool ServiceIntervalIsCustom() const;

	bool ServiceIntervalIsPercent() const;

	void SetServiceIntervalIsCustom(bool on);

	void SetServiceIntervalIsPercent(bool on);

	Order *GetOrder(int index) const;
	Order *GetLastOrder() const;

	bool IsEngineCountable() const;
	bool HasEngineType() const;
	bool HasDepotOrder() const;
	void HandlePathfindingResult(bool path_found);

	bool IsFrontEngine() const;
	bool IsArticulatedPart() const;
	bool HasArticulatedPart() const;
	Vehicle *GetNextArticulatedPart() const;
	Vehicle *GetFirstEnginePart();
	const Vehicle *GetFirstEnginePart() const;
	Vehicle *GetLastEnginePart();
	Vehicle *GetNextVehicle() const;
	Vehicle *GetPrevVehicle() const;
	IterateWrapper Orders() const;
};




/**
 * Class defining several overloaded accessors so we don't
 * have to cast vehicle types that often
 */
template <class T, VehicleType Type>
struct SpecializedVehicle : public Vehicle {
	static const VehicleType EXPECTED_TYPE = Type; ///< Specialized type

	typedef SpecializedVehicle<T, Type> SpecializedVehicleBase; ///< Our type

	/**
	 * Set vehicle type correctly
	 */
	SpecializedVehicle<T, Type>() : Vehicle(Type)
	{
		this->sprite_cache.sprite_seq.count = 1;
	}

	/**
	 * Get the first vehicle in the chain
	 * @return first vehicle in the chain
	 */
	T *First() const { return (T *)this->Vehicle::First(); }

	/**
	 * Get the last vehicle in the chain
	 * @return last vehicle in the chain
	 */
	T *Last() { return (T *)this->Vehicle::Last(); }

	/**
	 * Get the last vehicle in the chain
	 * @return last vehicle in the chain
	 */
	const T *Last() const { return (const T *)this->Vehicle::Last(); }

	/**
	 * Get next vehicle in the chain
	 * @return next vehicle in the chain
	 */
	T *Next() const { return (T *)this->Vehicle::Next(); }

	/**
	 * Get previous vehicle in the chain
	 * @return previous vehicle in the chain
	 */
	T *Previous() const { return (T *)this->Vehicle::Previous(); }

	/**
	 * Get the next part of an articulated engine.
	 * @return Next part of the articulated engine.
	 * @pre The vehicle is an articulated engine.
	 */
	T *GetNextArticulatedPart() { return (T *)this->Vehicle::GetNextArticulatedPart(); }

	/**
	 * Get the next part of an articulated engine.
	 * @return Next part of the articulated engine.
	 * @pre The vehicle is an articulated engine.
	 */
	T *GetNextArticulatedPart() const { return (T *)this->Vehicle::GetNextArticulatedPart(); }

	/**
	 * Get the first part of an articulated engine.
	 * @return First part of the engine.
	 */
	T *GetFirstEnginePart() { return (T *)this->Vehicle::GetFirstEnginePart(); }

	/**
	 * Get the first part of an articulated engine.
	 * @return First part of the engine.
	 */
	const T *GetFirstEnginePart() const { return (const T *)this->Vehicle::GetFirstEnginePart(); }

	/**
	 * Get the last part of an articulated engine.
	 * @return Last part of the engine.
	 */
	T *GetLastEnginePart() { return (T *)this->Vehicle::GetLastEnginePart(); }

	/**
	 * Get the next real (non-articulated part) vehicle in the consist.
	 * @return Next vehicle in the consist.
	 */
	T *GetNextVehicle() const { return (T *)this->Vehicle::GetNextVehicle(); }

	/**
	 * Get the previous real (non-articulated part) vehicle in the consist.
	 * @return Previous vehicle in the consist.
	 */
	T *GetPrevVehicle() const { return (T *)this->Vehicle::GetPrevVehicle(); }

	/**
	 * Tests whether given index is a valid index for vehicle of this type
	 * @param index tested index
	 * @return is this index valid index of T?
	 */
	static bool IsValidID(size_t index)
	{
		return Vehicle::IsValidID(index) && Vehicle::Get(index)->type == Type;
	}

	/**
	 * Gets vehicle with given index
	 * @return pointer to vehicle with given index casted to T *
	 */
	static T *Get(size_t index)
	{
		return (T *)Vehicle::Get(index);
	}

	/**
	 * Returns vehicle if the index is a valid index for this vehicle type
	 * @return pointer to vehicle with given index if it's a vehicle of this type
	 */
	static T *GetIfValid(size_t index)
	{
		return IsValidID(index) ? Get(index) : nullptr;
	}

	/**
	 * Converts a Vehicle to SpecializedVehicle with type checking.
	 * @param v Vehicle pointer
	 * @return pointer to SpecializedVehicle
	 */
	static T *From(Vehicle *v)
	{
		assert(v->type == Type);
		return (T *)v;
	}

	/**
	 * Converts a const Vehicle to const SpecializedVehicle with type checking.
	 * @param v Vehicle pointer
	 * @return pointer to SpecializedVehicle
	 */
	static const T *From(const Vehicle *v)
	{
		assert(v->type == Type);
		return (const T *)v;
	}

	/**
	 * Update vehicle sprite- and position caches
	 * @param force_update Force updating the vehicle on the viewport.
	 * @param update_delta Also update the delta?
	 */
	void UpdateViewport(bool force_update, bool update_delta)
	{
		bool sprite_has_changed = false;

		/* Skip updating sprites on dedicated servers without screen */
		if (_network_dedicated) return;

		/* Explicitly choose method to call to prevent vtable dereference -
		 * it gives ~3% runtime improvements in games with many vehicles */
		if (update_delta) ((T *)this)->T::UpdateDeltaXY();

		/*
		 * Only check for a new sprite sequence if the vehicle direction
		 * has changed since we last checked it, assuming that otherwise
		 * there won't be enough change in bounding box or offsets to need
		 * to resolve a new sprite.
		 */
		if (this->direction != this->sprite_cache.last_direction || this->sprite_cache.is_viewport_candidate) {
			VehicleSpriteSeq seq;

			((T*)this)->T::GetImage(this->direction, EIT_ON_MAP, &seq);
			if (this->sprite_cache.sprite_seq != seq) {
				sprite_has_changed = true;
				this->sprite_cache.sprite_seq = seq;
			}

			this->sprite_cache.last_direction = this->direction;
			this->sprite_cache.revalidate_before_draw = false;
		} else {
			/*
			 * A change that could potentially invalidate the sprite has been
			 * made, signal that we should still resolve it before drawing on a
			 * viewport.
			 */
			this->sprite_cache.revalidate_before_draw = true;
		}

		if (force_update || sprite_has_changed) {
			this->Vehicle::UpdateViewport(true);
		}
	}

	/**
	 * Returns an iterable ensemble of all valid vehicles of type T
	 * @param from index of the first vehicle to consider
	 * @return an iterable ensemble of all valid vehicles of type T
	 */
	static Pool::IterateWrapper<T> Iterate(size_t from = 0) { return Pool::IterateWrapper<T>(from); }
};



/** Generates sequence of free UnitID numbers */
struct FreeUnitIDGenerator {
	bool *cache;  ///< array of occupied unit id numbers
	UnitID maxid; ///< maximum ID at the moment of constructor call
	UnitID curid; ///< last ID returned; 0 if none

	FreeUnitIDGenerator(VehicleType type, CompanyID owner);
	UnitID NextID();

	/** Releases allocated memory */
	~FreeUnitIDGenerator() { free(this->cache); }
};



/** Sentinel for an invalid coordinate. */
static const int32 INVALID_COORD = 0x7fffffff;



void AddToShared(Vehicle *shared_chain, Vehicle *v_add);
void RemoveFromShared(Vehicle *v);
void LeaveStation(Vehicle *v);



#endif /* VEHICLE_BASE_H */
